/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  Copyright (c) 2000-2008 SpringCard - www.springcard.com
  See COPYING for copyright and legal info
  
  error_help.c
  -----------
  Low level access to a GemCore reader
  
*/

/*
 * This file is based on reader.c provided by GemPlus as part of the TLP
 * library
 */

/*****************************************************************************
*				Copyright (c) Gemplus Card International 1997
*
* Name		: reader.c
*
* Description : SmartCard reader interface

  Author	: JL GIRAUD
  Compiler	: Visual C++, GNU gcc
  Host		: Windows 9x, NT, Linux
*
* $Id: reader.c,v 1.11 2000/12/11 17:53:29 rousseau Exp $
*
******************************************************************************
*
* Warning	:
*
* Remarks	:
*
*****************************************************************************/

/*
 * SpringCard changes :
 * - Remove support for anything else than TLP
 * - Moved OS specific stuff to serial_palmos.c and serial_win32.c
 * - Added support for Unicode
 */

#include "springcard_i.h"

LONG GCRTranslateError(BYTE gcr_errcode)
{
  return 0;

}

TCHAR  *ISO7816_ErrorStr(BYTE sw1, BYTE sw2)
{
  static TCHAR iso7816_error[ISO7816_ERROR_TEXT_SIZE];
  iso7816_error[0] = '\0';

  if (sw1 == 0x90 && sw2 == 0x00)
  {
    _tcscat(iso7816_error, _T("Normal processing"));
    return iso7816_error;
  }

  if (sw1 == 0x61)
  {
    _stprintf(iso7816_error, _T("%d bytes of response still available."), sw2);
    return iso7816_error;
  }

  if (sw1 == 0x62)
  {
    _tcscat(iso7816_error, _T("State of non-volatile memory unchanged "));

    if (sw2 == 0x00)
      _tcscat(iso7816_error, _T("No informatin given."));
    if (sw2 == 0x81)
      _tcscat(iso7816_error, _T("Part of returned data my be corrupted."));
    if (sw2 == 0x82)
      _tcscat(iso7816_error, _T("End of file/record reached before reading Le bytes."));
    if (sw2 == 0x83)
      _tcscat(iso7816_error, _T("Selected file invalidated."));
    if (sw2 == 0x84)
      _tcscat(iso7816_error, _T("FCI not formated according to 5.1.5."));

    return iso7816_error;
  }

  if (sw1 == 0x63)
  {
    _tcscat(iso7816_error, _T("State of non-volatile memory changed. "));
    if (sw2 == 0x00)
      _tcscat(iso7816_error, _T("No information given."));
    if (sw2 == 0x81)
      _tcscat(iso7816_error, _T("File filled up by the last write."));
    if ((sw2 & 0xF0) == 0xC0)
    {
      TCHAR  *p = &iso7816_error[_tcslen(iso7816_error)];
      _stprintf(p, _T("Counter %d."), sw2 & 0x0F);
    }

    return iso7816_error;
  }

  if (sw1 == 0x64 && sw2 == 0x00)
  {
    _tcscat(iso7816_error, _T("State of non-volatile memory unchanged."));
    return iso7816_error;
  }

  if (sw1 == 0x65)
  {
    _tcscat(iso7816_error, _T("State of non-volatile memory changed. "));
    if (sw2 == 0x81)
      _tcscat(iso7816_error, _T("Memory failure."));
    return iso7816_error;
  }

  if (sw1 == 0x66)
  {
    _tcscat(iso7816_error, _T("Reserved for security-related issues."));
    return iso7816_error;
  }

  if (sw1 == 0x67)
  {
    _tcscat(iso7816_error, _T("Wrong length."));
    return iso7816_error;
  }

  if (sw1 == 0x68)
  {
    _tcscat(iso7816_error, _T("Functions in CLA not supported. "));
    if (sw2 == 0x81)
      _tcscat(iso7816_error, _T("Logical channel not supported."));
    if (sw2 == 0x82)
      _tcscat(iso7816_error, _T("Secure messaging not supported."));
    return iso7816_error;
  }

  if (sw1 == 0x69)
  {
    _tcscat(iso7816_error, _T("Command not allowed. "));
    if (sw2 == 0x81)
      _tcscat(iso7816_error, _T("Command incompatible with file structure."));
    if (sw2 == 0x82)
      _tcscat(iso7816_error, _T("Security status not satisfied."));
    if (sw2 == 0x83)
      _tcscat(iso7816_error, _T("Authentication methon blocked."));
    if (sw2 == 0x84)
      _tcscat(iso7816_error, _T("Referenced data invalidated."));
    if (sw2 == 0x85)
      _tcscat(iso7816_error, _T("Conditions of use not satisfied."));
    if (sw2 == 0x86)
      _tcscat(iso7816_error, _T("Command not allowed (no current EF)."));
    if (sw2 == 0x87)
      _tcscat(iso7816_error, _T("Expected SM data objects missing."));
    if (sw2 == 0x88)
      _tcscat(iso7816_error, _T("SM dta objects incorrect."));
    return iso7816_error;
  }

  if (sw1 == 0x6A)
  {
    _tcscat(iso7816_error, _T("Wrong parameter(s) P1-P2. "));
    if (sw2 == 0x80)
      _tcscat(iso7816_error, _T("Incorrect parameters in the data field."));
    if (sw2 == 0x81)
      _tcscat(iso7816_error, _T("Function not supported."));
    if (sw2 == 0x82)
      _tcscat(iso7816_error, _T("File not found."));
    if (sw2 == 0x83)
      _tcscat(iso7816_error, _T("Record not found."));
    if (sw2 == 0x84)
      _tcscat(iso7816_error, _T("Not enough memory space in the file."));
    if (sw2 == 0x85)
      _tcscat(iso7816_error, _T("Lc inconsistent with TLV structure."));
    if (sw2 == 0x86)
      _tcscat(iso7816_error, _T("Incorrect parameters P1-P2."));
    if (sw2 == 0x87)
      _tcscat(iso7816_error, _T("Lc inconsistent with P1-P2."));
    if (sw2 == 0x88)
      _tcscat(iso7816_error, _T("Referenced data not found."));
    return iso7816_error;
  }

  if (sw1 == 0x6B)
  {
    _tcscat(iso7816_error, _T("Wrong parameter(s) P1-P2."));
    return iso7816_error;
  }

  if (sw1 == 0x6C)
  {
    _stprintf(iso7816_error, _T("Wrong length Le: should be %d"), sw2);
    return iso7816_error;
  }

  if (sw1 == 0x6D)
  {
    _tcscat(iso7816_error, _T("Instruction code not supported or invalid."));
    return iso7816_error;
  }

  if (sw1 == 0x6E)
  {
    _tcscat(iso7816_error, _T("Class not supported."));
    return iso7816_error;
  }

  if (sw1 == 0x6F)
  {
    _tcscat(iso7816_error, _T("No precise diagnosis."));
    return iso7816_error;
  }

  /* Default error message */
  _stprintf(iso7816_error, _T("Not defined by 7816 (sw1=%02X, sw2=%02X)."), sw1, sw2);
  return iso7816_error;
}
