/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  Copyright (c) 2000-2008 SpringCard - www.springcard.com
  See COPYING for copyright and legal info

  pssc_help.c
  -----------
  String translation of PC/SC error codes

 */

#define _WINERROR_
#include "springcard_i.h"
#include "pcsc_impl.h"

const TCHAR *gcr_names[3] = { _T("GemCore"), _T("GemCore Lite"), _T("GemCore Lite Pro") };
BYTE gcr_family = 0xFF;

typedef struct
{
  LONG    code;
  TCHAR  *text;
}
SCARD_ERRORS_ST;

static const SCARD_ERRORS_ST SCARD_ERRORS[] = {
  {SCARD_S_SUCCESS, _T("No error was encountered.")},
  {SCARD_F_INTERNAL_ERROR, _T("An internal consistency check failed.")},
  {SCARD_E_CANCELLED, _T("The action was canceled by an SCardCancel request.")},
  {SCARD_E_INVALID_HANDLE, _T("The supplied handle was invalid.")},
  {SCARD_E_INVALID_PARAMETER, _T("One or more of the supplied parameters could not be properly interpreted.")},
  {SCARD_E_INVALID_TARGET, _T("Registry startup information is missing or invalid.")},
  {SCARD_E_NO_MEMORY, _T("Not enough memory available to complete this command.")},
  {SCARD_F_WAITED_TOO_LONG, _T("An internal consistency timer has expired.")},
  {SCARD_E_INSUFFICIENT_BUFFER, _T("The data buffer for returned data is too small for the returned data.")},
  {SCARD_E_UNKNOWN_READER, _T("The specified reader name is not recognized.")},
  {SCARD_E_TIMEOUT, _T("The user-specified timeout value has expired.")},
  {SCARD_E_SHARING_VIOLATION, _T("The smart card cannot be accessed because of other outstanding connections.")},
  {SCARD_E_NO_SMARTCARD, _T("The operation requires a smart card, but no smart card is currently in the device.")},
  {SCARD_E_UNKNOWN_CARD, _T("The specified smart card name is not recognized.")},
  {SCARD_E_CANT_DISPOSE, _T("The system could not dispose of the media in the requested manner.")},
  {SCARD_E_PROTO_MISMATCH, _T("The requested protocols are incompatible with the protocol currently in use with the card.")},
  {SCARD_E_NOT_READY, _T("The reader or card is not ready to accept commands.")},
  {SCARD_E_INVALID_VALUE, _T("One or more of the supplied parameter values could not be properly interpreted.")},
  {SCARD_E_SYSTEM_CANCELLED, _T("The action was canceled by the system, presumably to log off or shut down.")},
  {SCARD_F_COMM_ERROR, _T("An internal communications error has been detected.")},
  {SCARD_E_INVALID_ATR, _T("An ATR string obtained from the registry is not a valid ATR string.")},
  {SCARD_E_NOT_TRANSACTED, _T("An attempt was made to end a non-existent transaction.")},
  {SCARD_E_READER_UNAVAILABLE, _T("The specified reader is not currently available for use.")},
  {SCARD_E_PCI_TOO_SMALL, _T("The PCI receive buffer was too small.")},
  {SCARD_E_READER_UNSUPPORTED, _T("The reader driver does not meet minimal requirements for support.")},
  {SCARD_E_DUPLICATE_READER, _T("The reader driver didn't produce a unique reader name.")},
  {SCARD_E_CARD_UNSUPPORTED, _T("The smart card does not meet minimal requirements for support.")},
  {SCARD_E_NO_SERVICE, _T("The smart card resource manager is not running.")},
  {SCARD_E_SERVICE_STOPPED, _T("The smart card resource manager has shut down.")},
  {SCARD_E_UNEXPECTED, _T("An unexpected card error has occurred.")},
  {SCARD_E_UNSUPPORTED_FEATURE, _T("This smart card does not support the requested feature.")},
  {SCARD_E_NO_READERS_AVAILABLE, _T("None of the specified readers are currently available for use.")},
  {SCARD_E_COMM_DATA_LOST, _T("A communications error with the reader has been detected. Retry the operation.")},
  {SCARD_W_UNRESPONSIVE_CARD, _T("Card malfunction. The smart card is not responding to a reset or has aborded an exchange by timing out.")},
  {SCARD_W_UNPOWERED_CARD, _T("Power has been removed from the smart card, so that further communication is not possible.")},
  {SCARD_W_RESET_CARD, _T("The smart card has been reset, so any shared state information is invalid.")},
  {SCARD_W_REMOVED_CARD, _T("The smart card has been removed, so that further communication is not possible.")},
  {SCARD_F_UNKNOWN_ERROR, _T("An internal error has been detected, but the source is unknown.")},
  {SCARD_W_UNSUPPORTED_CARD, _T("The reader cannot communicate with the card, due to ATR string configuration conflicts.")}
};

TCHAR  *PCSC_ErrorStr(LONG err)
{
  static TCHAR errDefault[64];
  int     i;
  for (i = 0; i < sizeof(SCARD_ERRORS) / sizeof(SCARD_ERRORS_ST); i++)
  {
    if (SCARD_ERRORS[i].code == err)
      return SCARD_ERRORS[i].text;
  }
  _stprintf(errDefault, _T("Not a PC/SC error code (%d - 0x%08lX)"), (int) err, err);
  return errDefault;
}
