/* SpringCardS10API/S10_low
 *
 * NAME
 *   SpringCardS10API :: S10 Card low functions
 *
 * COPYRIGHT
 *   (c) 2009 SpringCard - www.springcard.com
 *
 * DESCRIPTION
 *   Implementation of S10 low functions to exchange data with the GemCore chipset
 *
 **/

#include "springcard_s10_i.h"

/* SpringCardS10API/SCardS10CardDirect
 *
 * NAME
 *   SCardS10CardDirect
 *
 * DESCRIPTION
 *   Raw exchange with the GemCore chipset
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE send_len     : length of data to send
 *   BYTE *send_data   : data to send
 *   BYTE recv_len     : length of data to receive
 *   BYTE *recv_data   : data received
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS   : success, card selected
 *   Other code if internal or communication error has occured
 *
 * SEE ALSO
 *   SCardS10CardTransmit
 *
 **/
LONG SCardS10CardDirect(SCARDHANDLE hCard, BYTE send_len, BYTE *send_data, BYTE recv_len, BYTE *recv_data, BYTE status[2])
{
  LONG  SCardLastError;
  BYTE  pbRecvBuffer[260];
  DWORD cbRecvLength;

  if (hCard == NULL)
    return SCARD_E_INVALID_HANDLE;

  if (send_data == NULL)
    return SCARD_E_INVALID_PARAMETER;

  cbRecvLength = sizeof(pbRecvBuffer);

  SCardLastError = SCardControl(hCard, send_data, send_len,	pbRecvBuffer, &cbRecvLength);
  if (SCardLastError != SCARD_S_SUCCESS) 
    return SCardLastError; 

  if (cbRecvLength < 2)
	return SCARD_E_INVALID_VALUE;

  if (status != NULL)
  {
    status[0] = pbRecvBuffer[cbRecvLength-2];        /* Status code SW1 */
    status[1] = pbRecvBuffer[cbRecvLength-1];        /* Status code SW2 */
  }

  if ((recv_len) && (recv_data != NULL)) 
  {
    if ( recv_len < cbRecvLength ) 
	  {
      cbRecvLength = recv_len;
	  }
	
	  memcpy(recv_data, pbRecvBuffer, cbRecvLength);
  }

  return SCARD_S_SUCCESS; 
}

/* SpringCardS10API/SCardS10CardTransmit
 *
 * NAME
 *   SCardS10CardTransmit
 *
 * DESCRIPTION
 *   Exchange APDU with the GemCore chipset
 *
 * INPUTS
 *   SCARDHANDLE hCard : handle of the card
 *   BYTE send_len     : length of data to send
 *   BYTE *send_data   : data to send
 *   BYTE recv_len     : length of data to receive
 *   BYTE *recv_data   : data received
 *   BYTE status[2]    : 2-byte buffer to receive status code
 *
 * RETURNS
 *   SCARD_S_SUCCESS   : success, card selected
 *   Other code if internal or communication error has occured
 *
 * SEE ALSO
 *   SCardS10CardTransmit
 *
 **/
LONG SCardS10CardTransmit(SCARDHANDLE hCard, BYTE send_len, BYTE *send_data, BYTE recv_len, BYTE *recv_data, BYTE status[2])
{
  LONG  SCardLastError;
  BYTE  pbRecvBuffer[260];
  DWORD cbRecvLength;

  if (hCard == NULL)
    return SCARD_E_INVALID_HANDLE;

  if (send_data == NULL)
    return SCARD_E_INVALID_PARAMETER;

  cbRecvLength = sizeof(pbRecvBuffer);

  SCardLastError = SCardTransmit(hCard, NULL, send_data, send_len, NULL, pbRecvBuffer, &cbRecvLength);
  if (SCardLastError != SCARD_S_SUCCESS)
    return SCardLastError; 

  if (cbRecvLength < 2)
	return SCARD_E_INVALID_VALUE;

  if ( status != NULL ) 
  {
    status[0] = pbRecvBuffer[cbRecvLength-2];        /* Status code SW1 */
    status[1] = pbRecvBuffer[cbRecvLength-1];        /* Status code SW2 */
  }

  if ((recv_len) && (recv_data != NULL)) 
  {
    if (recv_len < cbRecvLength) 
	  {
      cbRecvLength = recv_len;
	  }
	
    memcpy(recv_data, pbRecvBuffer, cbRecvLength);
  }

  return SCARD_S_SUCCESS; 
}
