/*
 *	Name : SpringCard_S9.c
 *	----------------------
 *	Source code for the SpringCard_S9 Dll
 *
 *	(c) Pro-Active S.A.S. - http://www.pro-active.fr
 *
 *	
 *	History :
 *		12/07/2007  LTC Creation
 *
 *---------------------------------------------------------------------------*/

#include "springcard_s9_i.h"


/*
 *	Global variables
 *	----------------
 *---------------------------------------------------------------------------*/
BOOL springcard_on   = FALSE;
BOOL springcard_card = FALSE;

SCARDCONTEXT SCardContext = NULL;
SCARDHANDLE SCardHandle   = NULL;
LONG SCardLastError;
DWORD SCardProtocol;
TCHAR SCardReaders[250];
TCHAR SCardErrorStr[250];

/*---------------------------------------------------------------------------*/
BOOL CardDirect(BYTE send_len, BYTE *send_data, BYTE recv_len, BYTE *recv_data, BYTE *status);
BOOL CardTransmit(BYTE send_len, BYTE *send_data, BYTE recv_len, BYTE *recv_data, BYTE *status);
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/*   Read Functions : Data Area , Protection Area , Security Area            */
/*---------------------------------------------------------------------------*/
BOOL S9ReadDataArea(BYTE *dest, WORD addr, BYTE len);
BOOL S9ReadProtectionArea(BYTE *dest, WORD addr);
BOOL S9ReadSecurityArea(BYTE *dest);

/*---------------------------------------------------------------------------*/
/*   Write Functions : Data Area , Protection Area , Security Area           */
/*---------------------------------------------------------------------------*/
BOOL S9WriteDataArea(BYTE *source, WORD addr, BYTE len);
BOOL S9WriteProtectionArea(BYTE *source, WORD addr, BYTE len);
BOOL S9UpdateSecurityArea(BYTE *source, WORD addr, BYTE len);

BOOL S9PresentCardSecretCode(BYTE code2, BYTE code1);



/*																STATUS : VALID
 *	Name : ReaderOpen
 *	-----------------
 *	Open the reader and retrieve its names (first name is the regular 
 *	smartcard slot, second name is the internal SIM/SAM slot).
 *---------------------------------------------------------------------------*/
BOOL ReaderOpen(void)
{
  DWORD l;

  if (!springcard_on)
  {
    /* Init PC/SC layer */
    SCardLastError = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &SCardContext);
    if (SCardLastError != SCARD_S_SUCCESS)
      goto error;

    /* Get readers list */
    l = sizeof(SCardReaders);
    SCardLastError = SCardListReaders(SCardContext, NULL, SCardReaders, &l);
    if (SCardLastError != SCARD_S_SUCCESS)
      goto error;

    /* OK ! */
    /* ---- */
    springcard_on = TRUE;
  }

  /* Display firmware */
  return TRUE;
  
error:
  return FALSE;
}

/*																STATUS : VALID
 *	Name : ReaderClose
 *	------------------
 *	Close the reader
 *---------------------------------------------------------------------------*/
BOOL ReaderClose(void)
{
  if (springcard_on)
  {
    CardClose();

    /* Free PC/SC layer */
    SCardReleaseContext(SCardContext);

    springcard_on = FALSE;
    return TRUE;
  }
  
  return FALSE;
}

/*																STATUS : VALID
 *	Name : CardOpen
 *	---------------
 *	Get access to a smartcard
 *---------------------------------------------------------------------------*/
BOOL CardOpen(void)
{
  /* SpringCard must be opened */
  if (!springcard_on) 
    return FALSE;

  if (!springcard_card)
  {
    SCardLastError = SCardConnect(SCardContext,
                                  SCardMultiStringItem(SCardContext, SCardReaders, 0),
                                  SCARD_SHARE_DIRECT,
                                  SCARD_PROTOCOL_RAW | SCARD_PROTOCOL_S9,
                                  &SCardHandle,
                                  &SCardProtocol);
    if (SCardLastError != SCARD_S_SUCCESS) 
      return FALSE;
	
    /* OK ! */
    /* ---- */
    springcard_card = TRUE;
  }

  return TRUE;
}

/*																STATUS : VALID
 *	Name : CardClose
 *	----------------
 *	Close the smartcard
 *---------------------------------------------------------------------------*/
BOOL CardClose(void)
{
  if (springcard_card)
  {
    /* Try to eject the card */
	  SCardDisconnect(SCardHandle, SCARD_UNPOWER_CARD);
	  springcard_card = FALSE;
	  return TRUE;
  }

  return FALSE;
}

/*																STATUS : VALID
 *	Name : CardDirect
 *	-----------------
 *	Raw exchange with the GemCore chipset
 *---------------------------------------------------------------------------*/
BOOL CardDirect(BYTE send_len, BYTE *send_data, BYTE recv_len, BYTE *recv_data, BYTE *status)
{
  if (!springcard_on)
	return FALSE;

  if (!springcard_card)
	return FALSE;

  SCardLastError = SCardS9CardDirect(SCardHandle, send_len, send_data, recv_len, recv_data, status);
  if (SCardLastError != SCARD_S_SUCCESS)
    return FALSE; 

  return TRUE; 
}
/*---------------------------------------------------------------------------*/
BOOL CardTransmit(BYTE send_len, BYTE *send_data, BYTE recv_len, BYTE *recv_data, BYTE *status)
{
  if (!springcard_on)
	return FALSE;

  if (!springcard_card)
	return FALSE;

  SCardLastError = SCardS9CardTransmit(SCardHandle, send_len, send_data, recv_len, recv_data, status);
  if (SCardLastError != SCARD_S_SUCCESS)
    return FALSE; 

  return TRUE; 
}

/*																STATUS : VALID
 *	Name : S9ReadDataArea
 *	----------------------
 *	S9 Read Function
 *---------------------------------------------------------------------------*/
BOOL S9ReadDataArea(BYTE *dest, WORD addr, BYTE len)
{ 
  BYTE status[2];

  SCardLastError = SCardS9ReadDataArea(SCardHandle, dest, addr, len, status);
  if (SCardLastError != SCARD_S_SUCCESS)
    return FALSE; 

  return TRUE; 
}

/*																STATUS : VALID
 *	Name : S9ReadProtectionArea
 *	----------------------------
 *	S9 Read Function
 *---------------------------------------------------------------------------*/
BOOL S9ReadProtectionArea(BYTE *dest, WORD addr)
{ 
  BYTE buffer[32];
  BYTE status[2];

  SCardLastError = SCardS9ReadProtectionArea(SCardHandle, buffer, addr, status);
  if (SCardLastError != SCARD_S_SUCCESS)
    return FALSE; 

  memcpy(dest, buffer, 32);

  return TRUE;
}

/*																STATUS : VALID
 *	Name : S9ReadSecurityArea
 *	--------------------------
 *	S9 Read Function
 *---------------------------------------------------------------------------*/
BOOL S9ReadSecurityArea(BYTE *dest)
{ 
  BYTE buffer[3];
  BYTE status[2];

  SCardLastError = SCardS9ReadSecurityArea(SCardHandle, buffer, status);
  if (SCardLastError != SCARD_S_SUCCESS)
    return FALSE; 

  memcpy(dest, buffer, 3);

  return TRUE;
}

/*																STATUS : VALID
 *	Name : S9WriteDataArea
 *	-------------
 *	S9 Write Function
 *---------------------------------------------------------------------------*/
BOOL S9WriteDataArea(BYTE *source ,WORD addr, BYTE len)
{ 
  BYTE status[2];

  SCardLastError = SCardS9WriteDataArea(SCardHandle, source, addr, len, status);
  if (SCardLastError != SCARD_S_SUCCESS)
    return FALSE; 

  return TRUE;
}

/*																STATUS : VALID
 *	Name : S9WriteProtectionArea
 *	-------------
 *	S9 Write Function
 *---------------------------------------------------------------------------*/
BOOL S9WriteProtectionArea(BYTE *source ,WORD addr, BYTE len)
{ 
  BYTE status[2];

  SCardLastError = SCardS9WriteProtectionArea(SCardHandle, source, addr, len, status);
  if (SCardLastError != SCARD_S_SUCCESS)
    return FALSE; 

  return TRUE;
}

/*																STATUS : VALID 
 *	Name : S9UpdateSecurityArea
 *	-------------
 *	S9 Write Function
 *---------------------------------------------------------------------------*/
BOOL S9UpdateSecurityArea(BYTE *source, WORD addr, BYTE len)
{ 
  BYTE status[2];

  SCardLastError = SCardS9UpdateSecurityArea(SCardHandle, source, addr, len, status);
  if (SCardLastError != SCARD_S_SUCCESS)
    return FALSE; 

  return TRUE;
}

/*																STATUS : VALID
 *	Name : S9PresentCardSecretCode
 *	-------------
 *	S9PresentCardSecretCode
 *---------------------------------------------------------------------------*/
BOOL S9PresentCardSecretCode(BYTE code2, BYTE code1)
{
  BYTE status[2];

  SCardLastError = SCardS9PSCVerification(SCardHandle, code2, code1, status);
  if (SCardLastError != SCARD_S_SUCCESS)
    return FALSE; 

  return TRUE; 
}
