/*****************************************************************************
*               Copyright (c) Gemplus Card International 1997
*
* Name        : serial.c
*
* Description : Homogenization of read-write function parameters

  Author      : L.GAUTERON
  Compiler    : Visual C++
  Host        : Windows 9x,NT
*
* $Id: serial.c,v 1.3 2000/12/11 17:53:30 rousseau Exp $
* $Log: serial.c,v $
* Revision 1.3  2000/12/11 17:53:30  rousseau
* Added a 5 seconds timeout if no reader present
*
* Revision 1.2  2000/09/04 08:19:06  rousseau
* code cleanup, changed the API and name for [Init|Close]Serial
*
*
******************************************************************************
*
* Warning     : 
*
* Remarks     : 
*
*****************************************************************************/

#ifdef __linux__

#define _DEBUG

#include "sctypes.h"
#include "reader.h"
#include "pcsc_impl.h"
#include "serial_linux.h"

#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <sys/file.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

TCHAR   sc_com_name[16] = "";
#define COM_NAME _T("/dev/ttyS%d")

int WriteCom(TYPE_HANDLE fd, LPCVOID lpBuffer, DWORD dwNbByte)
{
  return write(fd, lpBuffer, dwNbByte);
}

int ReadCom(TYPE_HANDLE fd, LPVOID lpBuffer, DWORD dwNbByte)
{
  fd_set  fdset;
  struct timeval timeout;
  int     res;

  /* timeout 2 seconds */
  timeout.tv_sec = 2;
  timeout.tv_usec = 0;

  FD_ZERO(&fdset);
  FD_SET(fd, &fdset);
  res = select(fd + 1, &fdset, NULL, NULL, &timeout);

  if (res)
    return read(fd, lpBuffer, dwNbByte);
  else
    return -1;
}

/*
 * GemPlus library function
 * ------------------------
 */
void SerialClose(ReaderDesc * ad_rd)
{
  tcsetattr(ad_rd->fd, TCSANOW, &(ad_rd->oldtio));
  lockf(ad_rd->fd, F_ULOCK, 0);
  close(ad_rd->fd);
}

int SerialOpen(ReaderDesc * ad_rd, TCHAR * device)
{
  struct termios newtio;
  int     rc;
  long    baudrate;

  ad_rd->fd = open(device, O_RDWR | O_NOCTTY);
  if (ad_rd->fd < 0)
  {
#ifdef _DEBUG
    perror(device);
#endif
    return -1;
  }

  /* test if the device is not used by an other process */
  if (lockf(ad_rd->fd, F_TLOCK, 0) == -1)
    return -1;

  tcgetattr(ad_rd->fd, &(ad_rd->oldtio)); /* save current port settings */

  bzero(&newtio, sizeof(newtio));
  /* BAUDRATE = 9600              */
  /* CS8  = 8n1 (8bit,no parity,1 stopbit */
  /* CLOCAL= local connection, no modem control */
  /* CREAD  = enable receiving characters */
  newtio.c_cflag = NORM_BAUDRATE | CS8 | CLOCAL | CREAD;
  newtio.c_iflag = IGNPAR;
  newtio.c_oflag = 0;

  /* set input mode (non-canonical, no echo,...) */
  newtio.c_lflag = 0;

  newtio.c_cc[VTIME] = 0;       /* inter-character timer unused */
  newtio.c_cc[VMIN] = 1;        /* blocking read until 1 chars received */

  tcflush(ad_rd->fd, TCIFLUSH);
  if (tcsetattr(ad_rd->fd, TCSANOW, &newtio))
  {
#ifdef _DEBUG
    perror("tcsetattr norm");
#endif
    close(ad_rd->fd);
    return -1;
  }
#ifdef GCR_HIGH_BAUD_RATE
  /* Ask the GemCore to enter high speed mode (38400bauds) */
  switch (HIGH_BAUDRATE)
  {
    case B1200:
      baudrate = 1200;
      break;
    case B2400:
      baudrate = 2400;
      break;
    case B4800:
      baudrate = 4800;
      break;
    case B9600:
      baudrate = 9600;
      break;
    case B19200:
      baudrate = 19200;
      break;
    case B38400:
      baudrate = 38400;
      break;
      //case B76800: baudrate= 76800; break;
    default:
      return -1;

  }
  rc = GBPSetSIOSpeed(ad_rd, baudrate);

  /* Don't care for the return code; change device speed to 38400 */
  newtio.c_cflag = HIGH_BAUDRATE | CS8 | CLOCAL | CREAD;
  if (tcsetattr(ad_rd->fd, TCSANOW, &newtio))
  {
#ifdef _DEBUG
    perror("tcsetattr high");
#endif
    close(ad_rd->fd);
    return -1;
  }
#endif

  return 0;
}

/*
 * Put reader online
 */
LONG RDR_OpenEx(S_CONTEXT_ST * context, TCHAR * port)
{
  LONG    rc;
  //printf("trying %s\n", port);
  if (SerialOpen(&context->desc, port) != -1)
  {
    rc = SCARD_GetFirmware(context);
    if (rc != SCARD_S_SUCCESS)
    {
      SerialClose(&context->desc);
      return rc;
    }
    /*
       printf("success !\n");    
       printf("%s\n", context->names);
       printf("%s\n", &context->names[strlen(context->names)+1]);
     */
    return rc;
  }
  return SCARD_E_INVALID_HANDLE;
}

LONG RDR_Open(S_CONTEXT_ST * context)
{
  DWORD   i;
  LONG    rc;

  if (context->enabled)
    return SCARD_S_SUCCESS;

  /* Try to reopen previously open port */
  /* ---------------------------------- */
  if (_tcslen(sc_com_name))
  {
    rc = RDR_OpenEx(context, sc_com_name);
    if (rc == SCARD_S_SUCCESS)
      goto ok;
  }

  /* Regular PC : browse up to 10 serial ports */
  /* ----------------------------------------- */
  for (i = 0; i <= 9; i++)
  {
    _stprintf(sc_com_name, COM_NAME, (int) i);
    rc = RDR_OpenEx(context, sc_com_name);
    if (rc == SCARD_S_SUCCESS)
      goto ok;
  }
  return SCARD_E_READER_UNAVAILABLE;  /* Reader not found, simply... */
ok:
  context->enabled = TRUE;
  context->start_count = 1;
  return SCARD_S_SUCCESS;
}

#endif
