/*
  THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.

  Copyright (c) 2000-2008 SpringCard - www.springcard.com
  See COPYING for copyright and legal info
  
  serial_visor.c
  ---------------
  Low level wrappers for RS232 communication on the Visor
  
*/

/*
 * This file is based on Ser166550.c, the Serial 16650 Library for
 * HandSpring Visor featuring a compatible UART on its SpringBoard
 * slot.
 */

#include "springcard_i.h"

#ifdef __palmos__
#ifndef POSE

#include "reader.h"

#include "serial_palmos.h"
#include "serial_visor.h"

#include <PalmOS.h>
#include <SystemPublic.h>
#include <SerialMgrOld.h>
#include <HsExt.h>

#define  NON_PORTABLE

/*
#include <SystemPrv.h>	  // to get task info for our patch
*/

/*
>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
 Copy these structures from the 3.0 SDK because they are no longer
  part of the new PalmOS SDK.

 IMPORTANT: SysKernelInfoSelector and SysKernelInfoType may only
  be used when running under PalmOS 3.1 since they are private to PalmOS
  and are not guaranteed to remain the same
*/

/* Selector codes */
typedef enum
{
  sysKernelInfoSelCurTaskInfo,
  sysKernelInfoSelTaskInfo,
  sysKernelInfoSelSemaphoreInfo,
  sysKernelInfoSelTimerInfo
}
SysKernelInfoSelector;

typedef struct SysKernelInfoType
{
  /*  <chg 1-Feb-1999 RM> Force to Byte for compiler independence */
  BYTE /*SysKernelInfoSelector */ selector; /* Which info to get */
  BYTE    filler;               /* padding */
  DWORD   id;                   /* which object to get info on */
  union
  {
    struct
    {
      DWORD   id;               /* ID of task                           */
      DWORD   nextID;           /* ID of next task                      */

      DWORD   tag;              /* task tag                             */
      DWORD   status;           /* task status                          */
      DWORD   timer;            /* # ticks left if task's timeout       */
      DWORD   timeSlice;        /* task's time slice                    */
      SWORD   priority;         /* task priority                        */
      WORD    attributes;       /* task attributes                      */
      SWORD   pendingCalls;     /* # of pending calls                   */
      DWORD   senderTaskID;     /* task ID of message sender            */
      DWORD   msgExchangeID;    /* task's message exchange ID, if any   */

      DWORD   tcbP;             /* pointer to TCB                       */
      DWORD   stackP;           /* stack pointer of task                */
      DWORD   stackStart;       /* top of task stack (where it started) */
      DWORD   stackSize;        /* size of stack                        */
    }
    task;

    struct
    {
      DWORD   id;
      DWORD   nextID;           /* ID of next semaphore                 */

      DWORD   tag;              /* semaphore tag                        */
      WORD    initValue;        /* semaphore initial value              */
      /*  >= 0 for counting, -1 for resource  */
      SWORD   curValue;         /* semaphore current value              */
      /*  >0 available                        */
      /*   0 not available                    */
      /*  -n not available, n tasks waiting   */
      SWORD   nestLevel;        /* 0 for counting, >= 0 for resource    */
      DWORD   ownerID;          /* owner ID                             */
      /*   0 for counting or resource free    */
    }
    semaphore;

    struct
    {
      DWORD   id;
      DWORD   nextID;           /* ID of next timer                     */

      DWORD   tag;              /* timer tag                            */
      DWORD   ticksLeft;        /* ticks left till timer runs           */
      DWORD   period;           /* timer period, if periodic            */
      DWORD   proc;             /* timer procedure                      */
    }
    timer;
  }
  param;

}
SysKernelInfoType;

typedef SysKernelInfoType *SysKernelInfoPtr;

/*
typedef void
PrvErrDisplayFileLineMsgFuncType (const Char * const filename,
								  UInt lineNo, const Char * const msg);

*/

static SER_GlobalsPtr gP = 0;

/*
 *  SER_ForHostID(void)
 *----------------------
 * Envoid le Host Id Compatibility
 */
HostID SER_ForHostID(void)
{
  return hostIDPalmOS;
}

/*
 * SER_WaitForChars
 * ----------------
 * Waits until the receive queue contains the specified 
 * number of characters or a timeout, whichever comes 
 * sooner.
 */
static WORD SER_WaitForChars(DWORD bytes, SDWORD timeout)
{
  WORD    err = 0;
  WORD    readyBytes, prevReadyBytes;
  SER_QueuePtr queP = &(gP->recvQ);
  BOOL    checkTimeout = FALSE;
  DWORD   endTime, currTime;
  SDWORD  waitTicks;

  /* Return right away if 0 bytes */
  if (!bytes)
    return 0;

  /* Caller must insure that bytes is less than the queue size */
  if (bytes >= queP->size)
    return serErrNoMem;

  /* If the App is using it's own wakeup handler, this call is invalid */
//  if (gP->wakeupHandler) return serErrTimeOut;

  /* Figure out the endTime and the timeout value to pass to the    */
  /* SysTaskWait() call. If timeout is infinite, the kernel wants 0 */
  /* If timeout is none (0) the kernel wants -1                     */

  currTime = TimGetTicks();     // init start of timeout countdown
  if (timeout < 0)
  {
    // Infinite wait
    timeout = 0;
    endTime = 0;
  } else if (timeout == 0)
  {
    // No wait
    timeout = -1;
    endTime = currTime;
    checkTimeout = true;
  } else
  {
    // Limited wait
    endTime = currTime + timeout;
  }

  /* Get the task ID of this task which will be (possibly) waiting for */
  /* characters                                                        */
  gP->taskID = SysTaskID();

  /* Instead of using up system resources for another semaphore,          */
  /* use SysTaskWait/SysTaskWake calls. The downside of these calls       */
  /* is that we may get "false" wake-ups, so we need to loop and re-check */
  /* the receive count every time we wake up.                             */
  prevReadyBytes = -1;
  while (1)
  {
    /* Calculate the # of bytes in the queue and break out */
    /* if there's enough                                   */
    readyBytes = (SWORD) queP->end - (SWORD) queP->start;

    if (readyBytes < 0)
      readyBytes += queP->size;
    if ((DWORD) readyBytes >= bytes)
      break;

    /* If first time through the loop, save the readyBytes */
    if (prevReadyBytes < 0)
      prevReadyBytes = readyBytes;

    /* Check for timeout if not first time through loop and not */
    /* infinite timeout (endTime==0)                            */
    currTime = TimGetTicks();
    if (checkTimeout && (endTime > 0) && (currTime >= endTime))
    {
      /* If we received any data since beginning of this timeout interval, */
      /* reset the timeout counter and keep waiting                        */
      if (timeout > 0 && readyBytes > prevReadyBytes)
      {
        /* Save the current ready byte count for a subsequent check */
        prevReadyBytes = readyBytes;

        /* Reset the timeout counter */
        endTime = currTime + timeout;
      }

      /* Otherwise, abort the wait with a "timeout" error */
      else
      {
        err = serErrTimeOut;
        break;
      }
    }

    /* Set semaphore threshold for the ISR. The ISR will call SysTaskWake() */
    /* when it detects enough characters in the queue to fill the request.  */
    gP->sigRecvUsedBytes = (WORD) bytes;

    /* If the bytes have arrived, break out.                                */
    /* This catches a potential race condition if the characters arrived    */
    /* after our first check but before we set the receive signal size.     */
    readyBytes = (SWORD) queP->end - (SWORD) queP->start;
    if (readyBytes < 0)
      readyBytes += queP->size;
    if ((DWORD) readyBytes >= bytes)
      break;

    /* Wait for the ISR to wake us up. If there is already a pending */
    /* wake for us, we will return from this call immediately.       */
    if (timeout > 0)
      waitTicks = endTime - currTime;
    else
      waitTicks = 0;            // infinite

    SysTaskWait(waitTicks);

    /* If there were any Serial errors, return with error code */
    if (gP->lineErrors)
    {
      err = serErrLineErr;
      break;
    }

    /* Check for timeout errs from now on */
    checkTimeout = true;

  }                             /* while(1) */

  /* No longer need signal from interrupt routine */
  gP->taskID = 0;

  /* Clear a possible extra Wake that may have sneaked in   */
  /* from the interrupt service routine before we had the   */
  /* chance to clear gP->taskID; this is necessary in order */
  /* to prevent side-effects on our client's timing logic   */

  SysTaskWaitClr();

  return err;
}

/*
 * SER_SendChars
 * -------------
 * Transmit the specified number of characters. Returns as soon
 * as characters are copied into transmit buffer. 
 */
static WORD SER_SendChars(BYTE * bufP, DWORD bytes, DWORD * bytesSentP)
{
  SER_QueuePtr queP = &(gP->sendQ);
  SDWORD  timeout = gP->settings.ctsTimeout;
  SDWORD  waitTicks;
  WORD    err = 0;
  DWORD   currTime;
  DWORD   endTime = 0;
  WORD    minSpace;
  BOOL    checkTimeout = FALSE;
  BYTE   *dstP;
  WORD    end;

  *bytesSentP = bytes;          /* init sent counter */

  /* Minimum amount of space in send buffer we need before we will */
  /* want to copy data into it                                     */
  minSpace = queP->size >> 1;

  /* Get the task ID of this task which will be (possibly) waiting for */
  /*  characters                                                       */
  gP->taskID = SysTaskID();

  /* Check for timeout only if it is non-infinite and auto-CTS is requested */
  if ((gP->settings.flags & serSettingsFlagCTSAutoM) && timeout >= 0)
    checkTimeout = true;

  /* Loop till all characters are copied into transmit queue */
  /* ------------------------------------------------------- */
  while (bytes)
  {
    WORD    space, chunk, prevSpace;
    SWORD   used;

    /* See how much space left in buffer. Note that we take the       */
    /*  size minus the used bytes minus 1. We have to subtract one    */
    /*  because we can't fit 'size' bytes into the queue - that would */
    /*  make the end offset equal the start offset which is true      */
    /*  when the queue is empty                                       */
    used = (SWORD) queP->end - (SWORD) queP->start;
    if (used < 0)
      used += queP->size;
    space = queP->size - used - 1;

    /* Figure out the endTime for this wait interval */
    /* --------------------------------------------- */
    if (checkTimeout)
    {
      currTime = TimGetTicks(); /* init start of timeout countdown */
      endTime = currTime + timeout;
    }

    /* If not enough space available, wait */
    /* ----------------------------------- */

    prevSpace = space;          /* save for timeout detection */

    if (bytes < minSpace)
      minSpace = (WORD) bytes;

    while (space < minSpace)
    {
      gP->sigSendFreeBytes = minSpace;

      /* check again for available space to avoid race condition */
      used = (SWORD) queP->end - (SWORD) queP->start;
      if (used < 0)
        used += queP->size;
      space = queP->size - used - 1;

      if (space >= minSpace)
        break;

      /* Check for timeout if not first time through loop and not */
      /* inifinite timeout (endTime==0)                           */
      currTime = TimGetTicks();
      if (checkTimeout && currTime >= endTime)
      {
        /* If transmit queue space changed since beginning of this       */
        /*  timeout interval, reset the timeout counter and keep waiting */
        if (timeout > 0 && space != prevSpace)
        {
          /* Save the current space for a subsequent check */
          prevSpace = space;

          /* Reset the timeout counter */
          endTime = currTime + timeout;
        }

        /* Otherwise, abort the wait with a "timeout" error */
        else
        {
          err = serErrTimeOut;
          break;
        }
      }

      /* Wait for the ISR to wake us up. If there is already a pending */
      /* wake for us, we will return from this call immediately.       */
      if (checkTimeout)
        waitTicks = endTime - currTime;
      else
        waitTicks = 0;          /* infinite */
      SysTaskWait(waitTicks);

    }                           /* If not enough space available, wait */

    /* If there were any errors, return with error code */
    if (gP->lineErrors)
      err = serErrLineErr;

    if (err)
      break;

    /* Copy the data into the transmit queue */
    /* ------------------------------------- */
    if (bytes > space)
      chunk = space;
    else
      chunk = (WORD) bytes;
    bytes -= chunk;

    dstP = queP->dataP + queP->end;
    end = queP->end;
    while (chunk--)
    {
      *dstP++ = *bufP++;
      if (end == queP->size - 1)
      {
        end = 0;
        dstP = queP->dataP;
      } else
        end++;

      queP->end = end;
    }

    /* Tell the hardware to start sending */
    SER_HwControl(gP, SER_HwCtlSendIntEnable, 0, 0);

  }                             /* while (bytes) */

  /* No longer need signal from interrupt routine */
  gP->taskID = 0;

  /* Clear a possible extra Wake that may have sneaked in   */
  /* from the interrupt service routine before we had the   */
  /* chance to clear gP->taskID; this is necessary in order */
  /* to prevent side-effects on our client's timing logic   */
  SysTaskWaitClr();

  /* Return number of bytes that were enqueued */
  *bytesSentP -= bytes;

  return err;
}

/*
 * SER_Open
 * ---------
 * Open the serial port
 */
WORD SER_Open(void)
{
  WORD    err;
  BYTE    removable;
  WORD    cardNo;

  /* Look for the first removable card */
  /* --------------------------------- */

  err = serErrBadPort;
  for (cardNo = 0; cardNo < MemNumCards(); cardNo++)
  {
    err = HsCardAttrGet(cardNo, hsCardAttrRemovable, &removable);
    if (err)
      goto Exit;

    if (removable)
      break;
  }

  /* If no removable card found, err */
  if (!removable)
  {
    err = serErrBadPort;
    goto Exit;
  }

  /* Allocate buffer */
  /* --------------- */

  if (!gP)
  {
    gP = MemPtrNew(sizeof(SER_GlobalsType));
    if (!gP)
    {
      err = memErrNotEnoughSpace;
      goto Exit;
    }
    MemPtrSetOwner(gP, 0);
    MemSet(gP, sizeof(SER_GlobalsType), 0);
  } else
    return serErrAlreadyOpen;

  /* Fill in default serial settings. */
  gP->settings.baudRate = SER_DefaultBaud;
  gP->settings.flags = SER_DefaultSettings;
  gP->settings.ctsTimeout = SER_DefaultCTSTimeout;

  /* Clear error code. */
  gP->lineErrors = 0;

  /* Allocate the receive buffer */
  /* --------------------------- */
  gP->recvQ.dataP = gP->recvQ.defDataP = MemPtrNew(SER_DefaultRecvQSize);
  if (!gP->recvQ.dataP)
  {
    err = memErrNotEnoughSpace;
    goto Exit;
  }

  MemPtrSetOwner(gP->recvQ.dataP, 0);

  gP->recvQ.start = 0;
  gP->recvQ.end = 0;
  gP->recvQ.size = SER_DefaultRecvQSize;

  /* Allocate the  send buffer */
  /* ------------------------- */
  gP->sendQ.dataP = gP->sendQ.defDataP = MemPtrNew(SER_DefaultSendQSize);
  if (!gP->sendQ.dataP)
  {
    err = memErrNotEnoughSpace;
    goto Exit;
  }

  MemPtrSetOwner(gP->sendQ.dataP, 0);
  gP->sendQ.start = 0;
  gP->sendQ.end = 0;
  gP->sendQ.size = SER_DefaultSendQSize;

  /* Init the hardware */
  /* ----------------- */

  /* This call will make sure the right card is installed */

  err = SER_HwInit(gP, cardNo);
  if (err)
    goto Exit;

  /* Set default settings */
  /* -------------------- */
  err = SER_SetSettings(&(gP->settings));

  /* Initialize the receive trigger threshold */
  gP->sigRecvUsedBytes = 1;

#if 0
  /* LA GROSSE CONNERIE POUR FORCER LA WAKE-UP IRQ QUI NE VIENT PAS TOUTE SEULE !!! */
  SER_HwControl(gP, SER_HwCtlSendIntEnable, 0, 0);
  if (!gP->irqCount)
  {
    gP->taskID = SysTaskID();
    SysTaskWait(2000);
    gP->taskID = 0;
  }
  /* Tout pt ou pas ? */

  ErrFatalDisplayIf(!gP->irqCount, "Bon, c'est tout pt");
#endif

Exit:

  /* Cleanup after errors */
  if (err)
  {
    if (gP)
      SER_Close();
  }

  return err;
}

/*
 * SER_Close
 * ---------
 * Closes the serial port
 */
WORD SER_Close(void)
{
  WORD    err;

  /* If not open, return err */
  if (!gP)
    return serErrNotOpen;

  /* If we're about to *really* close the serial manager, put the */
  /*  serial services to sleep                                    */
  SER_Sleep();

  /* Tell the hardware we're done */
  SER_HwFree(gP);

  /* Dispose of the receive and send queues */
  /* -------------------------------------- */
  if (gP->recvQ.defDataP)
  {
    err = MemPtrFree(gP->recvQ.defDataP);
    if (err)
      DbgBreak();
  }
  if (gP->sendQ.defDataP)
  {
    err = MemPtrFree(gP->sendQ.defDataP);
    if (err)
      DbgBreak();
  }

  /* Dispose of our globals data */
  /* --------------------------- */
  MemPtrFree(gP);
  gP = 0;

  return 0;
}

/*
 * SER_Sleep
 * ---------
 * Called by System when going to sleep or when the connection
 * is closed.
 */
WORD SER_Sleep(void)
{
  /* If not open, return */
  if (!gP)
    return 0;

  /* Put hardware to sleep */
  SER_HwSleep(gP);

  return 0;
}

/*
 * SER_Wake
 * --------
 * Called by System when Waking up
 */
WORD SER_Wake(void)
{
  /* If not open, return */
  if (!gP)
    return 0;

  /* Take hardware out of sleep */
  SER_HwWake(gP);

  return 0;
}

/*
 * SerGetSettings
 * --------------
 * Retreives settings of serial port.
 */
WORD SER_GetSettings(SER_SettingsPtr settingsP)
{
  /* If not open return err */
  if (!gP)
    return serErrNotOpen;

  settingsP = &(gP->settings);

  return 0;
}

/*
 * SerSetSettings
 * --------------
 * Changes settings of serial port.
 */
WORD SER_SetSettings(SER_SettingsPtr settingsP)
{
  WORD    err;
  WORD    valueLen;

  /* If not open return err */
  if (!gP)
    return serErrNotOpen;

  /* Setup the hardare appropriately */
  /* ------------------------------- */

  /* Flags */
  valueLen = sizeof(settingsP->flags);
  err = SER_HwControl(gP, SER_HwCtlFlagsSet, &settingsP->flags, &valueLen);
  if (err)
    goto Exit;

  /* Baud Rate */
  valueLen = sizeof(settingsP->baudRate);
  err = SER_HwControl(gP, SER_HwCtlBaudRateSet, &settingsP->baudRate, &valueLen);
  if (err)
    goto Exit;

  /* settingsP->ctsTimeout Timeout just gets saved in our globals */

  /* Save them in our globals */
  gP->settings = *settingsP;

Exit:
  return err;
}

/*
 * SER_GetStatusRI
 * -------------
 * Returns status of a serial connection.  The status returned
 * is compatible with RS-232 signaling.
 */
WORD SER_GetStatusRI(BOOL * riOnP)
{
  BYTE    on;
  WORD    valueLen;

  /* If not open return err */
  if (!gP)
    return serErrNotOpen;

  /* Query the hardware for the state of the cts and dsr inputs */
  /* ---------------------------------------------------------- */
  /* RI */
  valueLen = sizeof(on);
  SER_HwControl(gP, SER_HwCtlRI, &on, &valueLen);
  *riOnP = on;

  return gP->lineErrors;
}

/*
 * SER_GetStatus
 * -------------
 * Returns status of a serial connection.  The status returned
 * is compatible with RS-232 signaling.
 */
WORD SER_GetStatus(BOOL * ctsOnP, BOOL * dsrOnP)
{
  BYTE    on;
  WORD    valueLen;

  /* If not open return err */
  if (!gP)
    return serErrNotOpen;

  /* Query the hardware for the state of the cts and dsr inputs */
  /* ---------------------------------------------------------- */
  valueLen = sizeof(on);
  SER_HwControl(gP, SER_HwCtlCTS, &on, &valueLen);
  *ctsOnP = on;

  /* DSR */
  valueLen = sizeof(on);
  SER_HwControl(gP, SER_HwCtlDSR, &on, &valueLen);
  *dsrOnP = on;

  return gP->lineErrors;
}

/*
 * SER_ClearErr
 * ------------
 * Clears current error out of serial port.  Also reinitialized
 * the main IN and Out endpoints.
 */
WORD SER_ClearErr(void)
{
  /* If not open return err */
  if (!gP)
    return serErrNotOpen;

  /* Clear saved  error code */
  gP->lineErrors = 0;

  return 0;
}

/*
 * SER_Send
 * --------
 * Sends 1 or more bytes of data over the serial port.
 */
DWORD SER_Send(void *bufP, DWORD count, WORD * errP)
{
  WORD    err = 0;
  DWORD   numSent = 0;

  /* Make sure writable pointer arguments are valid */
  ErrFatalDisplayIf(!errP, "NULL ptr arg");

  /* If not open return err */
  if (!gP)
  {
    count = 0;
    err = serErrNotOpen;
    goto Exit;
  }

  /* Handle the transmit in the interrupt service routine, and signal */
  /* completion with a semaphore.                                     */
  err = SER_SendChars(bufP, count, &numSent);

Exit:
  *errP = err;

  return (numSent);
}

/*
 * SerSendWait
 * -----------
 * Waits for all characters in the send buffer to be sent
 * over the serial port
 */
WORD SER_SendWait(SDWORD timeout)
{
  SER_QueuePtr queP;
  DWORD   startTime, endTime;
  WORD    err = 0;
  BOOL    checkTimeout = FALSE;

  /* If not open return err */
  if (!gP)
    return serErrNotOpen;

  queP = &(gP->sendQ);

  /* If queue is empty, return right away */
  if (queP->end == queP->start)
    return 0;

  /* ------------------------------------------------------------- */
  /* Figure out the endTime and the timeout value to pass to the   */
  /*   SysTaskWait() call.                                         */
  /* If timeout is infinite, the kernel wants 0                    */
  /* If timeout is none (0) the kernel wants -1                    */
  /* ------------------------------------------------------------- */
  startTime = TimGetTicks();    /* init start of timeout countdown */
  if (timeout < 0)
  {
    timeout = 0;
    endTime = 0;
  } else if (timeout == 0)
  {
    timeout = -1;
    endTime = startTime;
    checkTimeout = TRUE;
  } else
    endTime = startTime + timeout;

  /* Get the task ID of this task which will be (possibly) waiting for */
  /*  characters                                                       */
  gP->taskID = SysTaskID();

  /* Desire entire buffer to be empty */
  gP->sigSendFreeBytes = queP->size;

  /* ------------------------------------------------------------- */
  /* Loop until transmit queue is empty                            */
  /* ------------------------------------------------------------- */
  do
  {
    if (queP->start == queP->end)
      break;

    /* Check for timeout if not first time through loop and not */
    /* inifinite timeout (endTime==0)                           */
    if (checkTimeout && endTime > 0 && TimGetTicks() >= endTime)
    {
      err = sysErrTimeout;
      break;
    }

    /* Wait for the ISR to wake us up. If there is already a pending */
    /* wake for us, we will return from this call immediately.       */
    SysTaskWait(timeout);
    checkTimeout = TRUE;

  } while (!err);

  /* Now, wait for the transmit FIFO to empty as well */
  if (!err)
    err = SER_HwSendFIFOWait(gP, checkTimeout, endTime);

  return err;
}

/*
 * SerSendCheck
 * ------------
 * Returns how many characters are left in the
 * transmit buffer
 */
WORD SER_SendCheck(DWORD * numBytesP)
{
  SER_QueuePtr queP;
  SWORD   bytes;

  /* If not open return err */
  if (!gP)
    return serErrNotOpen;

  queP = &(gP->sendQ);

  bytes = (SWORD) queP->end - (SWORD) queP->start;
  if (bytes < 0)
    bytes += queP->size;

  *numBytesP = bytes;
  return 0;
}

/*
 * SerSendFlush
 * ------------
 * Flushes characters out of the transmit queue.
 */
WORD SER_SendFlush(void)
{
  SER_QueuePtr queP;
  WORD    status;

  /* If not open return err */
  if (!gP)
    return serErrNotOpen;

  queP = &(gP->sendQ);

  /* Interrupts off */
  status = SysDisableInts();
  queP->start = queP->end = 0;
  SysRestoreStatus(status);

  /* NOTE: may also want to reset the transmit FIFO? */

  return 0;
}

/*
 * SerReceive
 * ----------
 * Receive the specified number of bytes  or returns with
 * error if a line error or timeout is encountered.
 */
DWORD SER_Receive(void *rcvBufP, DWORD count, SDWORD timeout, WORD * errP)
{
  BYTE   *bufP = (BYTE *) rcvBufP;
  DWORD   bytesLeft = count;    // save # of bytes to receive
  SER_QueuePtr queP;

  SWORD   maxChunkSize, chunk;
  register BYTE *srcP;
  register SWORD start;
  SWORD   maxStart;
  WORD    err = 0;

  /* Make sure writable ptr arguments are valid */
  ErrFatalDisplayIf((!errP || (!rcvBufP && count)), "NULL ptr arg");

  *errP = 0;                    /* init return error code */

  /* If not open return err */
  if (!gP)
  {
    err = serErrNotOpen;
    goto Exit;
  }

  queP = &(gP->recvQ);

  /* Loop on chunks that are smaller than the size of our queue */
  maxChunkSize = queP->size >> 1;
  while (bytesLeft)
  {
    /* If there are any serial errors, return with error code */
    if (gP->lineErrors)
    {
      err = serErrLineErr;
      break;
    }

    /* Figure out the chunk size */
    chunk = (SWORD) bytesLeft;
    if (chunk > maxChunkSize)
      chunk = maxChunkSize;

    /* Wait till we have enough bytes in the queue */
    err = SER_WaitForChars(chunk, timeout);
    if (err)
      break;

    /* Take out this chunk from the size */
    bytesLeft -= chunk;

    /* Transfer the data from the queue to the caller's buffer */
    start = queP->start;
    srcP = (BYTE *) queP->dataP + start;
    maxStart = queP->size - 1;
    while (chunk--)
    {
      *bufP++ = *srcP++;
      if (start == maxStart)
      {
        start = 0;
        srcP = (BYTE *) queP->dataP;
      } else
        start++;

      queP->start = start;
    }

    /* If the interrupt routine had filled the receive queue and disabled */
    /* interrupts, re-enable them now                                     */

    if (!err && gP->hw.needRecvIntEnable)
    {
      gP->hw.needRecvIntEnable = FALSE;
      SER_HwControl(gP, SER_HwCtlRecvIntEnable, 0, 0);
    }
  }                             /* while (bytesLeft) */

Exit:
  *errP = err;
  return (count - bytesLeft);
}

/*
 * SerReceiveChar
 * --------------
 * Receive one char, or returns with
 * error if a line error or timeout is encountered.
 */
WORD SER_ReceiveChar(BYTE * result, SDWORD timeout)
{
  WORD    err;
  SER_QueuePtr queP;

  /* If not open return err */
  if (!gP)
    return serErrNotOpen;

  queP = &(gP->recvQ);

  /* If there are any serial errors, return with error code */
  if (gP->lineErrors)
    return serErrLineErr;

  /* Wait till we have enough bytes in the queue */
  err = SER_WaitForChars(1, timeout);

  /* Get one char from the queue */
  *result = queP->dataP[queP->start];

  if (queP->start == queP->size - 1)
  {
    queP->start = 0;
  } else
  {
    queP->start++;
  }

  return err;
}

/*
 * SerReceiveWait
 * --------------
 * Wait for the specified number of bytes to arrive at the
 * serial port, or for a timeoout
 */
WORD SER_ReceiveWait(DWORD bytes, SDWORD timeout)
{
  WORD    err;
  SER_QueuePtr queP;

  /* If not open return err */
  if (!gP)
    return serErrNotOpen;

  queP = &(gP->recvQ);

  /* If number of bytes is bigger than our queue size, we can't handle */
  /* this call                                                         */
  if (bytes > queP->size)
  {
    return serErrBadParam;
  }

  /* If there are any serial errors, return with error code */
  if (gP->lineErrors)
    return serErrLineErr;

  /* Wait till we have enough bytes in the queue */
  err = SER_WaitForChars(bytes, timeout);

  return err;
}

/*
 * SER_ReceiveCheck
 * ----------------
 * Returns the number of characters currently in the
 * receive queue.
 */
WORD SER_ReceiveCheck(DWORD * numBytesP)
{
  SWORD   readyBytes;
  SER_QueuePtr queP;

  /* Assume none */
  *numBytesP = 0;

  /* If not open return err */
  if (!gP)
    return serErrNotOpen;

  /* Get the receive queue */
  queP = &(gP->recvQ);

  /* If there are any serial errors, return with error code */
  if (gP->lineErrors)
    return serErrLineErr;

  /* See how much is in the queue */
  readyBytes = (SWORD) queP->end - (SWORD) queP->start;
  if (readyBytes < 0)
    readyBytes += queP->size;

  *numBytesP = readyBytes;
  return 0;
}

/*
 * SerReceiveFlush
 * ---------------
 * Flushes all characters in the receive buffer and
 * all characters coming into the serial port until no
 * more arrive within the timeout period.  And clears the saved
 * error status.
 */
#define	prvMaxFlushErrorCount	50
void SER_ReceiveFlush(SDWORD timeout)
{
  BYTE    byteBuf;
  WORD    err;
  SWORD   errCount = 0;
  SER_QueuePtr queP;
  WORD    status;

  /* If not open return err */
  if (!gP)
    return;

  /* Get the receive queue * */
  queP = &(gP->recvQ);

  /* Interrupts off and flush existing chars */
  status = SysDisableInts();
  queP->start = queP->end = 0;
  SysRestoreStatus(status);

  /* If no timeout, return now */
  if (timeout == 0)
    return;

  while (true)
  {
    SER_Receive(&byteBuf, 1, timeout, &err);
    if (!err)
      continue;

    /* timeout */
    if (err == serErrTimeOut)
      break;

    /* Some other error */
    SER_ClearErr();             /* clear pending errors */
    errCount++;
    if (errCount > prvMaxFlushErrorCount)
      break;
  }

  SER_ClearErr();               /* clear pending errors */
}

/*
 * SER_SetReceiveBuffer
 * --------------------
 * Specify a new input buffer.
 */
WORD SER_SetReceiveBuffer(void *bufP, WORD bufSize)
{
  WORD    err = 0;
  WORD    sysStatus;
  SER_QueuePtr queP;

  /* If not open return err */
  if (!gP)
    return serErrNotOpen;

  queP = &(gP->recvQ);

  /* If replacing */
  if (bufSize)
  {
    sysStatus = SysDisableInts();
    queP->dataP = bufP;
    queP->size = bufSize;
    queP->start = queP->end = 0;
    SysRestoreStatus(sysStatus);

    /* Free the original buffer */
    if (queP->defDataP)
    {
      MemPtrFree(queP->defDataP);
      queP->defDataP = 0;
    }
  }

  /* Otherwise, restoring */
  /* -------------------- */
  else
  {
    /* Re-allocate the original receive buffer, if it was disposed of */
    if (!queP->defDataP)
    {
      queP->defDataP = MemPtrNew(SER_DefaultRecvQSize);
      if (queP->defDataP)
        MemPtrSetOwner(queP->defDataP, 0);
      else
      {
        FrmAlert(OutOfMemoryAlert);
        err = memErrNotEnoughSpace;
      }
    }

    sysStatus = SysDisableInts();
    queP->dataP = queP->defDataP;
    if (queP->dataP)
      queP->size = SER_DefaultRecvQSize;
    else
      queP->size = 0;
    queP->start = queP->end = 0;
    SysRestoreStatus(sysStatus);
  }

  return (err);
}

/*
 * SerControl
 * ----------
 */
WORD SER_Control(WORD op, void *valueP, WORD * valueLenP)
{
  WORD    err;

  /* Make sure serial lib has been opened for functions which require that */
  /* --------------------------------------------------------------------- */
  if (!gP && op != serCtlMaxBaud && op != serCtlHandshakeThreshold && op != serCtlIrScanningOn && op != serCtlIrScanningOff)
  {
    err = serErrNotOpen;
    goto Exit;
  }

  /* Process it now */
  /* -------------- */
  switch (op)
  {
    case serCtlMaxBaud:
      if (!valueP || !valueLenP || *valueLenP != sizeof(DWORD))
        err = serErrBadParam;
      else
      {
        *((DWORD *) valueP) = SER_MaxBaud;
        err = 0;
      }
      break;

    case serCtlHandshakeThreshold:
      if (!valueP || !valueLenP || *valueLenP != sizeof(DWORD))
        err = serErrBadParam;
      else
      {
        *((DWORD *) valueP) = SER_HandshakeThreshold;
        err = 0;
      }
      break;

      /* Pass the rest to the hardware layer to handle */
    default:
      err = SER_HwControl(gP, op, valueP, valueLenP);
      break;
  }

Exit:
  return (err);
}

void SER_LedOn(void)
{
  SER_HwLedOn(gP);
}

void SER_LedOff(void)
{
  SER_HwLedOff(gP);
}

int SER_EasyOpen(void)
{
  WORD    rc;
  SerSettingsType com_settings;

  /* Ouvre l'acces au port */
  rc = SER_Open();
  if ((rc != 0) && (rc != serErrAlreadyOpen))
  {
//    ShowMessage(MSG_ERROR + MSG_OK, SerialNoCardErrorString);
    return FALSE;
  }

  /* Rglages du port srie */
  com_settings.baudRate = 9600;
  com_settings.flags = (serSettingsFlagBitsPerChar8 | serSettingsFlagStopBits1 /*| serSettingsFlagRTSAutoM */ );
  com_settings.ctsTimeout = serDefaultCTSTimeout;

  if (SER_SetSettings(&com_settings))
  {
//    ShowMessage(MSG_ERROR + MSG_OK, SerialCommErrorString);
    SER_EasyClose();
    return FALSE;
  }

  /* Vidage du buffer d'mission */
  if (SER_SendFlush())
  {
//    ShowMessage(MSG_ERROR + MSG_OK, SerialCommErrorString);
    SER_EasyClose();
    return FALSE;
  }

  return TRUE;
}

void SER_EasyClose(void)
{
  SER_Close();
}

int SER_Read(BYTE * buffer, int max_buff_size)
{
  /* Variables locales  */
  unsigned int i = 0;
  unsigned int tries_count = 1;
  WORD    com_last_error = 0;

  while (i < max_buff_size)
  {
    com_last_error = SER_ReceiveChar((BYTE *) & (buffer[i]), 400);
    i++;
    if (com_last_error)
    {
      SER_ClearErr();
      return i;
    }
  }

  return i;
}

int SER_Write(BYTE * buffer, int buff_size)
{
  WORD    com_last_error;

  SER_ClearErr();
  SER_Send(buffer, buff_size, &com_last_error);
  if (!com_last_error)
    com_last_error = SER_SendWait(200);

  if (com_last_error != 0)
    return -1;
  else
    return buff_size;
}

// NOTE: The value of prvTxTrigBytes must be in sync with the FIFO Control
//  setting that is assigned to prvFifoCtlTxTrigSetting
#define prvTxTrigBytes			  8
#define prvFifoCtlTxTrigSetting	  SER_Hw16650FifoCtlTxTrig8

/***********************************************************************
 *
 * Function:     PrvSer16650IntHandler
 *
 * Description:  Card Interrupt handler for 16650 interrupts.
 *
 *	  All card interrupt handlers can be called from one of two possible
 *	  states of the device:
 *
 *		1.) System Awake. In this state, indicated by *sysAwakeP == true,
 *			all of the device hardware, except possibly the LCD, is 
 *			on and the interrupt handler is free to make any system calls
 *			that are normally safe to make from interrupt handlers. 
 *
 *			If the device came out of sleep mode due to this interrupt, 
 *			the default behavior is that it will go right back to 
 *			sleep again on the next loop through the current application's 
 *			event loop and leave the LCD off. This behavior permits an 
 *			interrupt to wake the device, perform limited processing 
 *			with the LCD off, and allow the device to go right back to 
 *			sleep again.  If instead, you want the device to stay awake 
 *			for a longer period of time or you want the LCD on (because 
 *			you're triggering an application to display something for 
 *			example), then you must call HsEvtResetAutoOffTimer() or 
 *			EvtResetAutoOffTimer().
 *
 *			HsEvtResetAutoOffTimer() lets you specify the minimum amount of 
 *			time to keep the device on before it goes back to deep sleep 
 *			mode and also lets you specify whether or not you want the
 *			LCD to turn on (if not already on). EvtResetAutoOffTimer()
 *			will keep the device on for the amount of time set by the
 *			user in the General Preferences panel and always turns on the
 *			LCD. 
 *
 *		2.) System not awake. In this state, indicated by *sysAwakeP == false,
 *			all of the device hardware, except for the CPU is still
 *			in sleep mode. The system will call the interrupt handler
 *			in this state if the card interrupt was the ONLY interrupt
 *			that occurred while the device was asleep. This allows the
 *			card interrupt handler to perform limited processing without
 *			incurring the overhead of the entire wake-up sequence. 
 *
 *			In this state, the interrupt handler MUST NOT MAKE ANY SYSTEM
 *			CALLS except for HsCardErrTry/HsCardErrCatch. If it does
 *			need to perform other system calls (like set a semaphore for
 *			example), then it should set the *sysAwakeP flag before it exits
 *			and return without clearing the interrupt source. The system will 
 *			then complete waking up the rest of the way (except for the LCD)
 *			and call the interrupt handler again with *sysWakeP set to true. 
 *			  
 *	  
 *
 * Parameters:   
 *	  cardParam		IN		Parameter to interrupt routine setup through
 *							the hsCardAttrCardParam attribute of HsCardAttrSet()
 *							In our case, this is a pointer to a structure
 *							of type HsDrcGlobalsType (defined in "HsDevRefCard.h"
 *							which has a pointer to the serial library globals
 *							inside of it. 
 *
 *	  *sysAwakeP	INOUT	This flag tells us if the system is fully awake
 *							or not. If this flag is false, the interrupt
 *							routine CAN NOT MAKE ANY SYSTEM CALLS except
 *							for HsCardErrTry/HsCardErrCatch. If the interrupt
 *							requires use of the system (to set a semaphore
 *							for example), then the handler
 *							should return without clearing the interrupt
 *							source and set the *sysAwakeP flag on exit.
 *
 *						  
 *               
 * Returns:      
 *	  error code, 0 if no error
 *
 * Called By:
 *	  System when card interrupt is asserted
 *
 * Notes:
 *
 *
 * History:
 *	  1-Mar-1999  RM	  Created
 *	  23-Sep-1999 VMK	  Removed unnecessary example code, added cases
 *						   for MSR and Xoff interrupts.
 *	  23-Sep-1999 VMK	  Disable RTS interrupts -- we don't need them;
 *						  Adjusted dependency on taskID being set to allow
 *						   the wakeup handler to be called when necessary;
 *						  Fixed to make sure that the wakeup handler gets
 *						   called only for receive-related events;
 *						  Clear the signal byte count when notifying;
 *						  Move check for receive signal threshold outside
 *						   character read loop -- performance optimization.
 *						  Move check for transmit signal threshold outside
 *						   character read loop -- performance optimization.
 *	  18-Nov-1999 VMK	  Fixed to take advantage of FIFO block transmit mode
 *
 ***********************************************************************/
#define prvISRTriggerLineError	  0x01
#define prvISRTriggerTX			  0x02
#define prvISRTriggerRX			  0x04
#define prvISRTriggerMSR		  0x08
#define prvISRTriggerRtsCts		  0x10
#define prvISRTriggerXoff		  0x20
#define prvISRTriggerOther		  0x40

static void SER_HwIntHandler(DWORD cardParam, BOOL * sysAwakeP)
{
  SER_Hw16650ReadType *readP;
  SER_Hw16650WriteType *writeP;
  WORD    lineStatus;
  WORD    end, start, freeBytes, fifoSlotsLeft;
  SWORD   usedBytes;
  BYTE   *dstP, *srcP;
  BYTE    status;
  BOOL    notifyApp = FALSE;
  BOOL    recvNotify = FALSE;   /* notify receiver */
  BOOL    sendNotify = FALSE;   /* notify sender   */

  /* Get the globals */
  SER_GlobalsPtr gP = (SER_GlobalsPtr) cardParam;

  ErrFatalDisplayIf(!gP, "Null UART globals");

  gP->irqCount += 1;

  /* Put a try/catch around this code that accesses the hardware */
  /* ----------------------------------------------------------- */
  HsCardErrTry
  {
    readP = (SER_Hw16650ReadType *) (gP->hw.uartP);
    writeP = (SER_Hw16650WriteType *) (gP->hw.uartP);

    SER_HwLedOn(gP);

    /* If no interrupt left to service, return */
    do
    {
      status = readP->intStatus;

      ErrFatalDisplayIf((status & 0xC0) != 0xC0, "FIFO's not enabled");

      /* If no more interrupt sources waiting to be serviced, bail out */
      if (status & 0x01)
      {

        break;
      }

      status &= 0x3E;

      /* Line error */
      /* ---------- */
      if (status == 0x06)       /* Line Status */
      {
        /* Get error */
        lineStatus = readP->lineStatus;
        if (lineStatus & SER_Hw16650LineStOverrun)
        {
          gP->lineErrors |= serLineErrorHWOverrun;
        }
        if (lineStatus & SER_Hw16650LineStParity)
          gP->lineErrors |= serLineErrorParity;
        if (lineStatus & SER_Hw16650LineStFraming)
          gP->lineErrors |= serLineErrorFraming;
        if (lineStatus & SER_Hw16650LineStBreak)
          gP->lineErrors |= serLineErrorBreak;

        /* Need to wake the task */
        notifyApp = recvNotify = true;
      }

      /* Line Status */
      /* Receive data ready */
      /* ------------------ */
      else if (status & 0x04)
      {
        /* case 0x04:   // Receive FIFO reached threshold */
        /* case 0x0C:   // Receive data timeout           */

        /* Get info on the receive queue */
        end = gP->recvQ.end;
        dstP = gP->recvQ.dataP + end;
        usedBytes = (SWORD) end - (SWORD) gP->recvQ.start;
        if (usedBytes < 0)
          usedBytes += gP->recvQ.size;

        while (readP->lineStatus & SER_Hw16650LineStReceive)
        {
          BYTE    data;

          /* Calculate next 'end' offset */
          end++;
          if (end == gP->recvQ.size)
            end = 0;

          /* If no space in receive buffer, disable further           */
          /*  interrupts and bail. The PrvWaitForChars() routine will */
          /*  re-enable receive interrupts when it empties out        */
          /*  the buffer                                              */
          if (end == gP->recvQ.start)
          {
            notifyApp = recvNotify = true;
            writeP->intEnable &= ~SER_Hw16650IntEnReceive;
            gP->hw.needRecvIntEnable = true;
            break;
          }

          /* Get the character out of the FIFO */
          data = readP->recData;

          /* Copy Char into buffer */
          *dstP++ = data;
          usedBytes++;

          /* Update end */
          gP->recvQ.end = end;
          if (end == 0)
            dstP = gP->recvQ.dataP;

        }                       /* while (readP->lineStatus & SER_Hw16650LineStReceive) */

        /* Wake task if signal reached */
        if (gP->sigRecvUsedBytes && usedBytes >= gP->sigRecvUsedBytes)
          notifyApp = recvNotify = true;

      }

      /* Receive data ready */
      /* Transmit FIFO has space */
      /* ----------------------- */
      else if (status == 0x02)
      {

        /* Get info on the send queue */
        start = gP->sendQ.start;
        srcP = gP->sendQ.dataP + start;
        usedBytes = (SWORD) gP->sendQ.end - (SWORD) start;
        if (usedBytes < 0)
          usedBytes += gP->sendQ.size;

        /* Subtract 1 because we can't fill the queue with size bytes */
        /* - that would make the end offset equal the start offset    */
        /* which is true when the queue is empty                      */
        freeBytes = gP->sendQ.size - usedBytes - 1;

        /* While FIFO has space */
        fifoSlotsLeft = SER_Hw16650TxFIFOSize - prvTxTrigBytes;
        do
        {
          BYTE    data;

          /* If no more in send buffer, disable further            */
          /* interrupts and bail. The PrvSendChars() routine will  */
          /* call SER_HwSendTrigger() to re-enable send interrupts */
          /* when it starts                                        */
          if (start == gP->sendQ.end)
          {
            notifyApp = sendNotify = true;
            writeP->intEnable &= ~SER_Hw16650IntEnTransmit;
            break;
          }

          /* Get the character out of the buffer and send */
          /* to FIFO                                      */
          data = *srcP++;
          writeP->sendData = data;

          /* Calculate next 'start' offset */
          start++;
          if (start == gP->sendQ.size)
          {
            start = 0;
            srcP = gP->sendQ.dataP;
          }

          /* Update queue */
          gP->sendQ.start = start;
          freeBytes++;

        }
        while (--fifoSlotsLeft);

        /* Wake task if signal reached */
        if (gP->sigSendFreeBytes && freeBytes >= gP->sigSendFreeBytes)
          notifyApp = sendNotify = true;

      }

      /* Transmit FIFO has space */
      /* MSR (Modem Status Register) */
      /* --------------------------- */
      else if (status == 0x0)
      {
        /* We should never get this interrupt because we don't */
        /* enable the Modem status interrupt                   */
      }

      /* MSR (Modem Status Register) */
      /* RTS/CTS hardware handshaking change of state */
      /* -------------------------------------------- */
      else if (status == 0x20)
      {
        /* Disable further handshake interrupts till queue gets emptied */
        /* PRESENTLY, WE DON'T ENABLE THE RTS INTERRUPT                 */
        /* writeP->intEnable &= ~SER_Hw16650IntEnRTS;                   */
        /* gP->hw.needRecvIntEnable = true;                             */
      }

      /* RTS/CTS hardware handshaking change of state */
      /* RXRDY (Received Xoff signal)/ Special character */
      /* ----------------------------------------------- */
      else if (status == 0x10)
      {
        /* We should never get this interrupt because we suppress */
        /* the software flow control features of the UART         */
      }

      /* RXRDY (Received Xoff signal)/ Special character */
      /* Some other unexpected interrupt status */
      /* -------------------------------------- */
      else
      {
      }
    } while (1);

    /* If we did something that requires notifying the application task,     */
    /* and the system is already awake enough to support system calls,       */
    /* notify the app now.  Note that the application can still get notified */
    /* about incoming characters and process them with the LCD off.          */
    if (notifyApp && *sysAwakeP)
    {
      /* If we have a receive-related notification... */
      if (recvNotify)
      {

        gP->sigRecvUsedBytes = 0;

        /* If there's a custom wakeup handler installed, call that */
//        if (gP->wakeupHandler)
//        {
//          SerWakeupHandler  procP;
//          procP = (SerWakeupHandler)gP->wakeupHandler;
//          (procP)(gP->wakeupRefcon);
//        }
//        else
        if (gP->taskID)
          SysTaskWake(gP->taskID);
      }

      /* If we have a transmit-related notification... */
      if (sendNotify)
      {
        gP->sigSendFreeBytes = 0;

        /* We don't call the wakeup handler in this case because it is */
        /*  for receive notifications only                             */

        /* If a task is waiting for our signal, wake it up now */

        if (gP->taskID)
          SysTaskWake(gP->taskID);
      }
    }
  }                             /* HsCardErrTry */

  /* If card got pulled, return error */
  /* -------------------------------- */
  HsCardErrCatch
  {
    ErrFatalDisplayIf(1, "Card got pulled during dialog, fatal exception !");
  }
HsCardErrEnd}

#define SILLY_CARDNO -1

/*
 * SER_HwInit
 * ----------
 * Initialize the hardware as a result of opening the Serial port
 */
WORD SER_HwInit(SER_GlobalsPtr gP, WORD cardNo)
{
  WORD    err = 0;
  DWORD   base, size;
  SER_Hw16650ReadType *readP;
  SER_Hw16650WriteType *writeP;
  DWORD   value;
#if 0
  char    cardName[32];
  BYTE    installed;
  WORD    version;
#endif

  /* Put a silly value in cardNo */
  gP->hw.cardNo = SILLY_CARDNO;

  /* Put a try/catch around this code that accesses the hardware */
  /* ----------------------------------------------------------- */
  HsCardErrTry
  {
#if 0
    /* If the card is not plugged in, return err */
    err = HsCardAttrGet(cardNo, hsCardAttrSwInstalled, &installed);
    ErrFatalDisplayIf(err, "Card software not installed");
    if (!installed)
    {
      err = serErrBadPort;
      goto DoneTry;
    }

    /* Make sure this is the correct card and correct version */
    err = MemCardInfo(cardNo, cardName, 0 /*manufName */ , &version,
                      0 /*crDateP */ , 0 /*romSizeP */ , 0 /*ramSizeP */ ,
                      0 /*freeBytesP */ );
    if (err)
      goto DoneTry;

    if (StrCompare(cardName, SER_HwCardName) || version < SER_HwCardVersion)
    {
      err = serErrBadPort;
      goto DoneTry;
    }
#endif

    /* Get the base address of the UART. It's addressed through csSlot1 so */
    /*  we take the CsBase and add CsSize to it                            */
    err = HsCardAttrGet(cardNo, hsCardAttrCsBase, &base);
    ErrFatalDisplayIf(err, "Unable to get card base address");

    err = HsCardAttrGet(cardNo, hsCardAttrCsSize, &size);
    ErrFatalDisplayIf(err, "Unable to get card size");

    gP->hw.uartP = (BYTE *) (base + size + SER_HwUartOffset);
    gP->hw.powerP = (BYTE *) (base + size + SER_HwPowerOffset);
    gP->hw.resetP = (BYTE *) (base + size + SER_HwResetOffset);

    /* Save the card number too */
    gP->hw.cardNo = cardNo;

    /* No waiting task... */
    gP->taskID = 0;

    /* Not any IRQ yet */
    gP->irqCount = 0;

    /* De-activate the card IRQ */
    value = (DWORD) 0;
    err = HsCardAttrSet(cardNo, hsCardAttrIntEnable, &value);
    ErrFatalDisplayIf(err, "Unable to de-activate card IRQ");

    /* Store our globals pointer into the card globals */
    value = (DWORD) gP;
    err = HsCardAttrSet(cardNo, hsCardAttrCardParam, &value);
    ErrFatalDisplayIf(err, "Unable to save globals pointer");

    /* Install our interrupt routine */
    /* ----------------------------- */
    value = (DWORD) SER_HwIntHandler;
    err = HsCardAttrSet(cardNo, hsCardAttrIntHandler, &value);
    ErrFatalDisplayIf(err, "Unable to activate card IRQ");

    /* Mark it as installed for SER_HwFree() */
    gP->hw.isrInstalled = TRUE;

    /* Power-up peripheral */
    /* ------------------- */
    SER_Hw5VOn(gP);

    /* Wait for power establishment and peripheral ready */
    /* ------------------------------------------------- */
    SysTaskDelay(SysTicksPerSecond() / 8);  /* JDA : 1/2 ou 1/4 de seconde ??? ca marche pas toujours !!! */

    /* Reset peripheral */
    /* ---------------- */
    *(gP->hw.resetP) = 0xFF;
    SysTaskDelay(SysTicksPerSecond() / 8);  /* JDA : a virer, ca fait chier */
    *(gP->hw.resetP) = 0x00;
    SysTaskDelay(SysTicksPerSecond() / 8);  /* JDA : a virer, ca fait chier */

    /* Configure the UART chip now */
    /* --------------------------- */
    readP = (SER_Hw16650ReadType *) (gP->hw.uartP);
    writeP = (SER_Hw16650WriteType *) (gP->hw.uartP);

    /* Set int registers to default values */
    writeP->intEnable = 0;

    /* Hardware test */
    writeP->scratch = 0x75;
    ErrFatalDisplayIf(readP->scratch != 0x75, "Card hardware error !");

    /* Enable enhanced features */
    {
      SER_Hw16650EnhType *enhP = (SER_Hw16650EnhType *) writeP;
      WORD    intsStatus;

      /* Disable interrupts while we set the control register to     */
      /* a special value, in order to avoid undesirable side-effects */
      /* (such as getting stuck in an infinite loop in the interrupt */
      /* handler)                                                    */
      intsStatus = SysDisableInts();

      writeP->lineControl = SER_Hw16650LineCtlEnFeatures;
      enhP->enhReg = SER_Hw16650EnhEnableAdv; /* Enable enhanced features */
      writeP->lineControl = 0;

      /* Restore interrupts */
      SysRestoreStatus(intsStatus);
    }

    /* Zero out again to clear enhanced bits */
    writeP->intEnable = 0;

    /* Turn on DMA mode (bit 3) and enable FIFOs (bit 0)                      */
    /* (all commands written to the FIFO control register must always have    */
    /*  the FIFO Enable bit set; otherwise, the FIFO mode will be turned off) */
    gP->hw.reg.fifoControl = SER_Hw16650FifoCtlDMA + SER_Hw16650FifoCtlFIFOEnable;

    /* Set the TX FIFO triggier level (bits 4 & 5) to interrupt the processor  */
    /* when the # of bytes in the transmit FIFO drops below a given # of bytes */
    /* (this reduces the number of interrupts the processor has to service,    */
    /* improving throughput, system performance, and battery life)             */
    gP->hw.reg.fifoControl |= prvFifoCtlTxTrigSetting;

    /* Set the RX FIFO trigger level (bits 6 & 7) to interrupt the processor   */
    /* when the # of bytes in the receive FIFO equals 24                       */
    gP->hw.reg.fifoControl |= SER_Hw16650FifoCtlRxTrig24;

    /* Now, write the accumulated value to the FIFO Control register */
    /* (FIFOs must be enabled before writing any other bits          */
    /* in the fifoControl register)                                  */
    writeP->fifoControl = SER_Hw16650FifoCtlFIFOEnable;
    writeP->fifoControl = gP->hw.reg.fifoControl;

    /* Line control to 0 default */
    writeP->lineControl = 0;

    /* Activate the IRQx outputs (bit 3) since we're using PC mode */
    /* of the chip                                                 */
    writeP->modemControl = SER_Hw16650ModemCtlIRQx;

    /* Reset the FIFOs */
    writeP->fifoControl = (gP->hw.reg.fifoControl | 0x06);

    /* We jump to here if an error occurs so that we can clean up */
    /* the HsCardErrTry                                           */

  DoneTry:
    ;
  }

  /* If card got pulled, return error */
  /* -------------------------------- */
  HsCardErrCatch
  {
    ErrFatalDisplayIf(1, "Card got pulled during init, fatal exception !");
  }
  HsCardErrEnd return err;
}

/*
 * SER_HwFree
 * ----------
 * Undo whatever SER_HwInit does
 */
WORD SER_HwFree(SER_GlobalsPtr gP)
{
  WORD    err = 0;
  SER_Hw16650WriteType *writeP = 0;
  DWORD   value;

  if (gP->hw.cardNo != SILLY_CARDNO)
  {
    /* De-activate the card IRQ */
    /* ------------------------ */
    value = (DWORD) 0;
    err = HsCardAttrSet(gP->hw.cardNo, hsCardAttrIntEnable, &value);

    /* Remove our interrupt routine */
    /* ---------------------------- */
    value = (DWORD) 0;
    HsCardAttrSet(gP->hw.cardNo, hsCardAttrIntHandler, &value);

    /* Zero out our globals pointer that we stored in the card globals */
    /* for our interrupt routine.                                      */
    value = 0;
    err = HsCardAttrSet(gP->hw.cardNo, hsCardAttrCardParam, &value);
    if (err)
      goto Exit;

    /* Put a try/catch around this code that accesses the hardware */
    /* ----------------------------------------------------------- */
    HsCardErrTry
    {
      /* Power-down peripheral */
      /* --------------------- */
      SER_Hw5VOff(gP);

      /* Initialize the UART chip now */
      /* ---------------------------- */
      writeP = (SER_Hw16650WriteType *) (gP->hw.uartP);

      /* Set all registers to default values */
      writeP->intEnable = 0;
      writeP->fifoControl = gP->hw.reg.fifoControl = 0;
      writeP->lineControl = 0;
      writeP->modemControl = 0;
    }

    /* If card got pulled, return error */
    /* -------------------------------- */
    HsCardErrCatch
    {
      err = serErrBadPort;
    }
  HsCardErrEnd}

Exit:
  ErrFatalDisplayIf(err && err != serErrBadPort, "Error in HwFree");
  return err;
}

/*
 * SER_HwControl
 * -------------
 * Set or retrieve hardware parameters or modes of the UART. 
 */
WORD SER_HwControl(SER_GlobalsPtr gP, WORD op, void *valueP, void *valueLenP)
{
  WORD    err = 0;
  SER_Hw16650WriteType *writeP = (SER_Hw16650WriteType *) (gP->hw.uartP);
  SER_Hw16650ReadType *readP = (SER_Hw16650ReadType *) (gP->hw.uartP);
  BYTE    lineControl;
  BYTE    retry = 10;
  BOOL    riOnP;

  /* Put a try/catch around this code that accesses the hardware */
  /* ----------------------------------------------------------- */
  HsCardErrTry
  {
    switch (op)
    {
      case serCtlStartBreak:
        writeP->lineControl |= SER_Hw16650LineCtlBreakSet;
        gP->hw.breakOn = true;
        break;

      case serCtlStopBreak:
        writeP->lineControl &= ~SER_Hw16650LineCtlBreakSet;
        gP->hw.breakOn = false;
        break;

      case serCtlBreakStatus:
        *((WORD *) valueP) = gP->hw.breakOn;
        break;

      case serCtlStartLocalLoopback:
        writeP->modemControl |= SER_Hw16650ModemCtlLoopback;
        break;

      case serCtlStopLocalLoopback:
        writeP->modemControl &= ~SER_Hw16650ModemCtlLoopback;
        break;

      case serCtlIrDAEnable:
        /* This depends on the enhanced feature set being enabled (see EFR bit-4) */
        writeP->modemControl |= SER_Hw16650ModemCtlIREnable;
        break;

      case serCtlIrDADisable:
        /* This depends on the enhanced feature set being enabled (see EFR bit-4) */
        writeP->modemControl &= ~SER_Hw16650ModemCtlIREnable;
        break;

      case serCtlIrScanningOn:
        err = serErrNotSupported;
        break;

      case serCtlIrScanningOff:
        err = serErrNotSupported;
        break;

        /* rxEnable/disable used for half duplex IrDA comm */
      case serCtlRxEnable:
        err = serErrNotSupported;
        break;

      case serCtlRxDisable:
        err = serErrNotSupported;
        break;

      case SER_HwCtlCTS:
        if (readP->modemStatus & SER_Hw16650ModemStCTS)
          *((BYTE *) valueP) = TRUE;
        else
          *((BYTE *) valueP) = FALSE;
        break;

      case SER_HwCtlDSR:
        if (readP->modemStatus & SER_Hw16650ModemStDSR)
          *((BYTE *) valueP) = TRUE;
        else
          *((BYTE *) valueP) = FALSE;
        break;

      case SER_HwCtlRI:
        if (readP->modemStatus & SER_Hw16650ModemStRI)
          *((BYTE *) valueP) = TRUE;
        else
          *((BYTE *) valueP) = FALSE;
        break;

        /* Sent by PrvWaitChars() after emptying receive queue somewhat */
      case SER_HwCtlRecvIntEnable:
        /* Re-enable receive interrupts */
        while (retry)
        {
          writeP->intEnable |= /*SER_Hw16650IntEnRTS | */ SER_Hw16650IntEnReceive;
          if (writeP->intEnable & SER_Hw16650IntEnReceive)
            break;
          retry--;
        }
        ErrFatalDisplayIf(!retry, "Unable to set RX-IRQ");
        break;

        /* Sent by PrvSendChars() after putting data into send queue */
      case SER_HwCtlSendIntEnable:
        /* Re-enable send interrupts */
        while (retry)
        {
          writeP->intEnable |= SER_Hw16650IntEnTransmit;
          if (writeP->intEnable & SER_Hw16650IntEnTransmit)
            break;
          retry--;
        }
        ErrFatalDisplayIf(!retry, "Unable to set TX-IRQ");
        break;

      case SER_HwCtlBaudRateSet:
        {
          DWORD   baud;
          WORD    reg;
          SER_Hw16650BaudType *baudP;

          baud = *((DWORD *) valueP);

          /* Check for max */
          if (baud > SER_MaxBaud)
          {
            err = serErrNotSupported;
            break;
          }

          /* Figure out what divisor we need. The generated baud rate */
          /* is the crystal frequency / 16 / baudReg.   */
          /* We test RI pin to determine the Quartz of the Springcard Module */
          /* RI = TRUE (1) ==> Quartz = 7372800                              */
          /* RI = TRUE (0) ==> Quartz = 14745600                             */

          SER_GetStatusRI(&riOnP);
          if (!riOnP)
            reg = (WORD) ((SER_HwClockFreq1 / 16 + baud / 2) / baud);
          else
            reg = (WORD) ((SER_HwClockFreq2 / 16 + baud / 2) / baud);

          /* Err if reg is 0 */
          if (reg == 0)
          {
            err = serErrNotSupported;
            break;
          }

          {
            WORD    intsStatus;

            /* Disable interrupts while we modify multiple registers to */
            /*  avoid undesirable side-effects (such as getting stuck   */
            /*  in an infinite loop in the interrupt handler)           */
            intsStatus = SysDisableInts();

            /* Turn off the divide by 4 logic */
            writeP->modemControl &= ~SER_Hw16650ModemCtlClockSel;

            /* Set up to access the baud rate registers */
            writeP->lineControl |= SER_Hw16650LineCtlDivLatchEn;

            baudP = (SER_Hw16650BaudType *) writeP;
            baudP->baudLSB = reg & 0x0FF;
            baudP->baudMSB = reg >> 8;

            /* Back to normal */
            writeP->lineControl &= ~SER_Hw16650LineCtlDivLatchEn;

            /* Set flag */
            gP->hw.baudInited = true;

            /* Restore interrupts */
            SysRestoreStatus(intsStatus);
          }
        }
        break;

        /* Flags */
        /* ----- */
      case SER_HwCtlFlagsSet:
        {
          DWORD   flags;
          SER_Hw16650EnhType *enhP;

          flags = *((DWORD *) valueP);

          /* Set stop bits */
          /* ------------- */
          switch (flags & serSettingsFlagStopBitsM)
          {
            case serSettingsFlagStopBits1:
              writeP->lineControl &= ~SER_Hw16650LineCtlStop;
              break;
            case serSettingsFlagStopBits2:
              writeP->lineControl |= SER_Hw16650LineCtlStop;
              break;
          }

          /* Set Parity */
          /* ---------- */
          writeP->lineControl &= ~(SER_Hw16650LineCtlParitySet | SER_Hw16650LineCtlParityEven | SER_Hw16650LineCtlParityEn);
          if (flags & serSettingsFlagParityOnM)
            writeP->lineControl |= SER_Hw16650LineCtlParityEn;

          if (flags & serSettingsFlagParityEvenM)
            writeP->lineControl |= SER_Hw16650LineCtlParityEven;

          /* Ignore Software flow control for now... */
          /* --------------------------------------- */
          if (flags & serSettingsFlagXonXoffM)
            err = serErrNotSupported;

          /* RTS and CTS receive flow control */
          /* -------------------------------- */
          enhP = (SER_Hw16650EnhType *) writeP;

          /* Set line control to 0xBF to access extended features */
          /* ---------------------------------------------------- */
          {
            WORD    intsStatus;

            /* Disable interrupts while we set the control register to     */
            /* a special value, in order to avoid undesirable side-effects */
            /* (such as getting stuck in an infinite loop in the interrupt */
            /* handler)                                                    */
            intsStatus = SysDisableInts();

            lineControl = writeP->lineControl;
            writeP->lineControl = SER_Hw16650LineCtlEnFeatures;

            ErrFatalDisplayIf(!(enhP->enhReg & SER_Hw16650EnhEnableAdv), "Enhanced features not enabled");
            if (flags & serSettingsFlagRTSAutoM)
              enhP->enhReg |= SER_Hw16650EnhAutoRTS;
            else
              enhP->enhReg &= ~SER_Hw16650EnhAutoRTS;

            if (flags & serSettingsFlagCTSAutoM)
              enhP->enhReg |= SER_Hw16650EnhAutoCTS;
            else
              enhP->enhReg &= ~SER_Hw16650EnhAutoCTS;

            writeP->lineControl = lineControl;

            /* Restore the interrupts */
            SysRestoreStatus(intsStatus);
          }

          /* Bits per character */
          /* ------------------ */
          writeP->lineControl &= ~(SER_Hw16650LineCtlData1 | SER_Hw16650LineCtlData0);
          switch (flags & serSettingsFlagBitsPerCharM)
          {
            case serSettingsFlagBitsPerChar5:
              writeP->lineControl |= 0;
              break;
            case serSettingsFlagBitsPerChar6:
              writeP->lineControl |= SER_Hw16650LineCtlData0;
              break;
            case serSettingsFlagBitsPerChar7:
              writeP->lineControl |= SER_Hw16650LineCtlData1;
              break;
            case serSettingsFlagBitsPerChar8:
              writeP->lineControl |= SER_Hw16650LineCtlData0 | SER_Hw16650LineCtlData1;
              break;
          }

          /* Set flag */
          gP->hw.modeInited = TRUE;
        }
        break;

        /* Return err if we don't recognize it */
      default:
        err = serErrBadParam;
        break;
    }

    /* If we just finished initializing the chip, enable interrupts */
    /* ------------------------------------------------------------ */
    if (!gP->hw.initFinished && gP->hw.baudInited && gP->hw.modeInited)
    {
      BYTE    enable;

      /* Turn on DTR and RTS and the LED */
      writeP->modemControl |= SER_Hw16650ModemCtlRTS | SER_Hw16650ModemCtlDTR;

      /* Enable the UART receive and line status (err) interrupts */
      while (retry)
      {
        writeP->intEnable |= SER_Hw16650IntEnRcvStatus | SER_Hw16650IntEnReceive;
        if (writeP->intEnable & (SER_Hw16650IntEnRcvStatus | SER_Hw16650IntEnReceive))
          break;
        retry--;
      }
      ErrFatalDisplayIf(!retry, "You will never get any  interrupt !");

      /* Enable card interrupts to the motherboard */
      enable = TRUE;
      err = HsCardAttrSet(gP->hw.cardNo, hsCardAttrIntEnable, &enable);

      if (!err)
        gP->hw.initFinished = TRUE;
    }
  }                             /* HsCardErrTry */

  /* If card got pulled, return error */
  /* -------------------------------- */
  HsCardErrCatch
  {
    err = serErrBadPort;
  }
  HsCardErrEnd ErrNonFatalDisplayIf(err && err != serErrBadPort, "SER_HwControl err");
  return (err);
}

/*
 * SER_HwSleep
 * -----------
 * Put Hardware into low power sleep mode
 */
WORD SER_HwSleep(SER_GlobalsPtr gP)
{
  WORD    err = 0;
  SER_Hw16650WriteType *writeP;

  /* Put a try/catch around this code that accesses the hardware */
  /* ----------------------------------------------------------- */
  HsCardErrTry
  {
    writeP = (SER_Hw16650WriteType *) (gP->hw.uartP);

    /* NOTE: May also want to disable interrupts, because the 650's             */
    /* low power state may still trigger interrupts if data keeps arriving      */
    /* in the serial port (such as when the device is turned off, but left      */
    /* connected to the data source). May also need to disable the line driver, */
    /* if any is used in the product.  Investigate this...                      */

    /* Put into low power mode. */

    /* This depends on the enhanced feature set being enabled (see EFR bit-4) */
    writeP->intEnable |= SER_Hw16650IntEnSleep; /* low power bit */
  }

  /* If card got pulled, return error */
  /* -------------------------------- */
  HsCardErrCatch
  {
    err = serErrBadPort;
  }
  HsCardErrEnd ErrNonFatalDisplayIf(err && err != serErrBadPort, "Error in HwSleep");
  return err;
}

/*
 * SER_HwWake
 * ----------
 * Take Hardware out of low power sleep mode
 */
WORD SER_HwWake(SER_GlobalsPtr gP)
{
  WORD    err = 0;
  SER_Hw16650WriteType *writeP;

  /* Put a try/catch around this code that accesses the hardware */
  /* ----------------------------------------------------------- */
  HsCardErrTry
  {
    writeP = (SER_Hw16650WriteType *) (gP->hw.uartP);

    /* Take out of low power mode */
    writeP->intEnable &= ~SER_Hw16650IntEnSleep;  /* low power bit */
  }

  /* If card got pulled, return error */
  /* -------------------------------- */
  HsCardErrCatch
  {
    err = serErrBadPort;
  }
  HsCardErrEnd ErrNonFatalDisplayIf(err && err != serErrBadPort, "SER_HwWake err");
  return err;
}

/*
 * SER_HwSendFIFOWait
 * ------------------
 */
WORD SER_HwSendFIFOWait(SER_GlobalsPtr gP, BOOL checkTimeout, SDWORD ctsTimeout)
{
  WORD    err = 0;
  SER_Hw16650ReadType *readP;
  SDWORD  waitStartTicks;

  ErrNonFatalDisplayIf(checkTimeout && (ctsTimeout < 0), "Negative ctsTimeout");

  /* Put a try/catch around this code that accesses the hardware */
  /* ----------------------------------------------------------- */
  HsCardErrTry
  {
    readP = (SER_Hw16650ReadType *) (gP->hw.uartP);

    /* Loop till fifo empties */
    do
    {
      /* Check if TX FIFO is empty */
      if (readP->lineStatus & SER_Hw16650LineStTransmit)
        break;

      /* If using CTS timeout and CTS is OFF, wait for it to turn on */
      /*  while checking for timeout                                 */
      if (checkTimeout && !(readP->modemStatus & SER_Hw16650ModemStCTS))
      {
        waitStartTicks = TimGetTicks();

        /* Wait for CTS to toggle */
        do
        {
          /* Check for CTS toggle */
          if (readP->modemStatus & SER_Hw16650ModemStDeltaCTS)
            break;

          /* Check again if TX FIFO is empty (just in case) */
          if (readP->lineStatus & SER_Hw16650LineStTransmit)
            break;

          /* See if timeout yet */
          if ((TimGetTicks() - waitStartTicks) >= ctsTimeout)
            err = serErrTimeOut;
        }
        while (!err);
      }
    } while (!err);
  }

  /* If card got pulled, return error */
  /* -------------------------------- */
  HsCardErrCatch
  {
    err = serErrBadPort;
  }
  HsCardErrEnd
    //ErrNonFatalDisplayIf (err && err!=serErrBadPort, "SER_HwSendFIFOWait err");
    return err;
}

/*
 * 5V Power
 * --------
 */
void SER_Hw5VOn(SER_GlobalsPtr gP)
{
  HsCardErrTry
  {
    *(gP->hw.powerP) = 0xFF;
  }
  HsCardErrCatch
  {
  }
HsCardErrEnd}

void SER_Hw5VOff(SER_GlobalsPtr gP)
{
  HsCardErrTry
  {
    *(gP->hw.powerP) = 0x00;
  }
  HsCardErrCatch
  {
  }
HsCardErrEnd}

/*
 * UART LED
 * --------
 */

void SER_HwLedOff(SER_GlobalsPtr gP)
{
  HsCardErrTry
  {
    ((SER_Hw16650WriteType *) (gP->hw.uartP))->modemControl &= ~0x04;
  }
  HsCardErrCatch
  {
  }
HsCardErrEnd}

void SER_HwLedOn(SER_GlobalsPtr gP)
{
  HsCardErrTry
  {
    ((SER_Hw16650WriteType *) (gP->hw.uartP))->modemControl |= 0x04;
  }
  HsCardErrCatch
  {
  }
HsCardErrEnd}

#endif
#endif
