#include "stdafx.h"
#include "I2C.h" 

#include "Inc/sctypes.h"
#include "Inc/springcard.h"

BOOL springcard_on   = FALSE;
BOOL springcard_card = FALSE;

SCARDCONTEXT SCardContext = -1;
SCARDHANDLE SCardHandle   = -1;
LONG SCardLastError;
DWORD SCardProtocol;
TCHAR SCardReaders[250];
TCHAR SCardErrorStr[250];

BOOL mutex = FALSE;

#ifdef _DEBCUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/*----------------------------------------------------------*/
/*
 * Close the smartcard
 */
void CardClose(void)
{
  if (springcard_card)
  {
    /* Try to eject the card ;-) */
    SCardDisconnect(SCardHandle, SCARD_UNPOWER_CARD);   
       
    springcard_card = FALSE;
  }  
}
/*----------------------------------------------------------*/
/*
 * Close the reader
 */
void ReaderClose(void)
{
  if (springcard_on)
  {    
    CardClose();
    
    /* Free PC/SC layer */
    SCardReleaseContext(SCardContext);
      
    springcard_on = FALSE;
  }
}
/*----------------------------------------------------------*/
/*
 * Open the reader and retrieve its names (first name is the regular smartcard slot,
 * second name is the internal SIM/SAM slot
 */
BOOL ReaderOpen(void)
{
  DWORD l;
  
  if (!springcard_on)
  {
    /* Init PC/SC layer */
    SCardLastError = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &SCardContext);
    if (SCardLastError != SCARD_S_SUCCESS)
      goto error;
    
    /* Get readers list */
    l = sizeof(SCardReaders);
    SCardLastError = SCardListReaders(SCardContext, NULL, SCardReaders, &l);
    if (SCardLastError != SCARD_S_SUCCESS)
      goto error;
   
    /* OK ! */  
    springcard_on = TRUE;
  }
    
  /* Display firmware */      
  return TRUE;
  
error:
  return FALSE;
}
/*----------------------------------------------------------*/
/*
 * Get access to a smartcard
 */
BOOL CardOpen(void)
{
  /* SpringCard must be opened */
  if (!springcard_on)
    return FALSE;

  if (!springcard_card)
  {
    SCardLastError = SCardConnect(SCardContext,
                                  SCardMultiStringItem(SCardContext, SCardReaders, 0),
                                  SCARD_SHARE_DIRECT,
                                  SCARD_PROTOCOL_RAW,
                                  &SCardHandle,
                                  &SCardProtocol);

    if (SCardLastError != SCARD_S_SUCCESS)
      return FALSE;  

    /* OK ! */
    springcard_card = TRUE;
  }

  return TRUE;
}
//-------------------------------------------------------------------------
BOOL CardRead(BYTE *dest, BYTE addr_hi, BYTE addr_low,BYTE len)
{
  BOOL can_retry = TRUE;

  if (mutex) return FALSE;
  mutex = TRUE;


	CWaitCursor wait;
	wait.Restore();

retry:
  if (!ReaderOpen())
	goto failed;

  if (!CardOpen())
    goto failed;

  if (!Card8KI2CRead(dest, addr_hi , addr_low, len))
    goto failed;

  // Close Card
  CardClose();

  // Close Reader
  ReaderClose();

  goto leave;

failed:
  //	m_err = CString(SCardErrorToString(SCardContext, SCardLastError));
  ReaderClose();
  if (can_retry)
  {
    can_retry = FALSE;
    goto retry;
  }
  mutex = FALSE;
  return FALSE;
  

leave:
  mutex = FALSE;
  return TRUE;
}
//-------------------------------------------------------------------------
BOOL CardWrite(BYTE *source, BYTE addr_hi, BYTE addr_low,BYTE len)
{
  BOOL can_retry = TRUE;

  if (mutex) return FALSE;
  mutex = TRUE;


	CWaitCursor wait;
	wait.Restore();

retry:
  if (!ReaderOpen())
	goto failed;

  if (!CardOpen())
    goto failed;

  if (!Card8KI2CWrite(source, addr_hi , addr_low, len))
    goto failed;

  // Close Card
  CardClose();

  // Close Reader
  ReaderClose();

  goto leave;

failed:
  //	m_err = CString(SCardErrorToString(SCardContext, SCardLastError));
  ReaderClose();
  if (can_retry)
  {
    can_retry = FALSE;
    goto retry;
  }
  mutex = FALSE;
   return FALSE;
 

leave:
  mutex = FALSE;
  return TRUE;
}
/*----------------------------------------------------------*/
/*
 * Raw exchange with the GemCore chipset
 */
BOOL CardDirect(BYTE send_len, BYTE *send_data, BYTE recv_len, BYTE *recv_data, BYTE *status)
{
  BYTE  pbRecvBuffer[260];
  DWORD cbRecvLength;

  if (!springcard_on) return FALSE;
  if (!springcard_card) return FALSE;

  cbRecvLength = sizeof(pbRecvBuffer),
 
  SCardLastError = SCardControl(SCardHandle,
                                send_data,
                                send_len,
                                pbRecvBuffer,
                                &cbRecvLength
                               );

  if (SCardLastError != SCARD_S_SUCCESS)
  {
    return FALSE; 
  }

  if (cbRecvLength < 2)
    return FALSE;

  if (status != NULL)
  {
    status[0] = pbRecvBuffer[cbRecvLength-2];        /* Status code SW1 */
    status[1] = pbRecvBuffer[cbRecvLength-1];        /* Status code SW2 */
  }

  if ((recv_len) && (recv_data != NULL))
  {
    if (recv_len < cbRecvLength) cbRecvLength = recv_len;
    memcpy(recv_data, pbRecvBuffer, cbRecvLength);
  }

  return TRUE; 
}

/****************************************************************************
 *                                                                          *
 *          GemCore Lite low-level calls for synchronous smartcards         *
 *                                                                          *
 ****************************************************************************/

static BYTE POWER_UP_CMD[] =
{
  0x16, // Synchronous interpreter command

  0x00, // A  = CLA    don't care
  0x12, // R5 = INS    don't care
  0x00, // R6 = P1     don't care
  0x00, // R7 = P2     don't care
  0x00, // R2 = Lin    don't care
  0x00, // R3 = Lout   don't care
  0x02, // 2 lines to follow

  0x11, //       %VCC_ON
  0x12  //       %RESET
};

static BYTE S8K_I2C_READ_CMD[] =
{
  0x16, // Synchronous interpreter command

  0x00, // A  = CLA    don't care
  0xB0, // R5 = INS    don't care
  0x00, // R6 = P1   = offset H 
  0xCC, // R7 = P2   = offset L
  0x00, // R2 = Lin    don't care
  0xCC, // R3 = Lout = size to read
  33,   // lines to follow

  0x21, //       %CLR_RST             ' reset = 0

// Start Condition
  0x51, //       %SET_IO              ' SDA = 1
  0x71, //       %SET_CLK             ' CLK = 1 
  0x41, //       %CLR_IO              ' SDA = 0

  0x74, //       MOV      A,0xA0      ' A = 0xA0 (write command)
  0xA0,
  0x92, //       %WRH_L               ' write A

// Word Adress (1)
  0x71, //       %SET_CLK             ' CLK = 1
  0x51, //       %SET_IO              ' SDA = 0

  0xEE, //       MOV      A,R6        ' A = R6 = P1 (read offset Hi)
  0x92, //       %WRH_L               ' write A

// Word Adress (0)
  0x71, //       %SET_CLK             ' CLK = 1
  0x51, //       %SET_IO              ' SDA = 0

  0xEF, //       MOV      A,R7        ' A = R7 = P2 (read offset low)
  0x92, //       %WRH_L               ' write A

// Ack ?
  0x71, //       %SET_CLK             ' clock = 1 
  0x61, //       %CLR_CLK             ' clock = 0

// Start Condition
  0x51, //       %SET_IO              ' SDA = 1
  0x71, //       %SET_CLK             ' CLK = 1 
  0x41, //       %CLR_IO              ' SDA = 0

  0x74, //       MOV      A,0xA1      ' A = 0xA1 (read command)
  0xA1,
  0x92, //       %WRH_L               ' write A

        // loop:

  0x41, //       %CLR_IO              ' io = 0
  0x71, //       %SET_CLK             ' clock = 1
  0x61, //       %CLR_CLK             ' clock = 0

  0x43, //       %RDH_L               ' A = read
  0xF6, //       MOV      @R0,A       ' *R0 = A
  0x08, //       INC      R0          ' R0++
  0xDB, //       DJNZ     R3,0xF8     ' R3--, if R3<>0 goto loop (R3 = L_out !)
  0xF8,
  0xA2, //       %RST_PUL             ' reset card
  0x42  //       %RET_OK              ' data_out = [R4..R0], sw1=0x90, sw2=0x00
};

/************************/
/* 8K I2C Read Function */
/************************/
BOOL Card8KI2CRead(BYTE *dest, BYTE addr_hi, BYTE addr_low,BYTE len)
{
  BYTE buffer[130];
  BYTE status[2];


  if (len > 128 ) return FALSE;

  /* Power-up (again ?) the card */
  if (!CardDirect(sizeof(POWER_UP_CMD), POWER_UP_CMD, sizeof(buffer), buffer, status))
    return FALSE;


  S8K_I2C_READ_CMD[3] = addr_hi;
  S8K_I2C_READ_CMD[4] = addr_low;
  S8K_I2C_READ_CMD[6] = len;

  if (!CardDirect(sizeof(S8K_I2C_READ_CMD), S8K_I2C_READ_CMD, sizeof(buffer), buffer, status))
    return FALSE;

  memcpy(dest, buffer, len);
 
  return TRUE; 
}

static BYTE S8K_I2C_WRITE_CMD[] =
{
  0x16, // Synchronous interpreter command

  0x00, // A  = CLA    don't care
  0xB0, // R5 = INS    don't care
  0x00, // R6 = P1   = offset H 
  0x00, // R7 = P2   = offset L
  32  , // R2 = Lin    Data in
  0x54,	// Data 001	  // T //Adress in R1 
  0x68,	// Data 002	    h
  0x65,	// Data 003	    e
  0x20,	// Data 004		
  0x71,	// Data 005		q
  0x75,	// Data 006		u
  0x69,	// Data 007		i	
  0x63,	// Data 008		c	
  0x6B,	// Data 009		k
  0x20,	// Data 010

  0x62,	// Data 011		b
  0x72,	// Data 012		r		
  0x6F,	// Data 013		o
  0x77,	// Data 014		w	
  0x6E,	// Data 015		n
  0x20,	// Data 016
  0x66,	// Data 017		f
  0x6F,	// Data 018		o
  0x78,	// Data 019		x
  0x20,	// Data 020

  0x6A,	// Data 021		j
  0x75,	// Data 022		u
  0x6D,	// Data 023		m
  0x70,	// Data 024		p
  0x73,	// Data 025		s
  0x20,	// Data 026
  0x6F,	// Data 027		o
  0x76,	// Data 028		v	
  0x65,	// Data 029		e							
  0x72,	// Data 030		r

  0x20,	// Data 031
  0x74,	// Data 032		t

  32, // R3 = Lout = size to write
  29,   // lines to follow

  0x21, //       %CLR_RST             ' reset = 0

// Start Condition
  0x51, //       %SET_IO              ' SDA = 1
  0x71, //       %SET_CLK             ' CLK = 1 
  0x41, //       %CLR_IO              ' SDA = 0

  0x74, //       MOV      A,0xA0      ' A = 0xA0 (write command)
  0xA0,
  0x92, //       %WRH_L               ' write A

// Ack ?
  0x71, //       %SET_CLK             ' CLK = 1
  0x51, //       %SET_IO              ' SDA = 1

// Word Adress (1)
  0xEE, //       MOV      A,R6        ' A = R6 = P1 (read offset Hi)
  0x92, //       %WRH_L               ' write A

// Ack ?
  0x71, //       %SET_CLK             ' CLK = 1
  0x51, //       %SET_IO              ' SDA = 1

// Word Adress (0)
  0xEF, //       MOV      A,R7        ' A = R7 = P2 (read offset low)
  0x92, //       %WRH_L               ' write A

        // loop:
// Ack ?
  0x51, //       %SET_IO              ' SDA = 1
  0x71, //       %SET_CLK             ' CLK = 1
  

  0xE7, //       MOV      A,@R1       ' *R1 -> A
  0x92, //       %WDH_L               ' A write
  0x09, //       INC      R1          ' R1++
  0xDB, //       DJNZ     R3,        ' R3--, if R3<>0 goto loop (R3 = L_out !)
  0xF9,

// Stop Condition
  0x71, //       %SET_CLK             ' CLK = 1
  0x61, //       %CLR_CLK             ' CLK = 0
  0x41, //       %CLR_IO              ' SDA = 0
  0x71, //       %SET_CLK             ' CLK = 1
  0x51, //       %SET_IO              ' SDA = 1

  0xA2, //       %RST_PUL             ' reset card
  0x42  //       %RET_OK              ' data_out = [R4..R0], sw1=0x90, sw2=0x00
};

/*************************/
/* 8K I2C Write Function */
/*************************/
BOOL Card8KI2CWrite(BYTE *source, BYTE addr_hi, BYTE addr_low,BYTE len)
{
  BYTE buffer[130];
  BYTE status[2];

  if (len > 32 ) return FALSE;

  /* Power-up (again ?) the card */
  if (!CardDirect(sizeof(POWER_UP_CMD), POWER_UP_CMD, sizeof(buffer), buffer, status))
    return FALSE;


  S8K_I2C_WRITE_CMD[3] = addr_hi;
  S8K_I2C_WRITE_CMD[4] = addr_low;

  S8K_I2C_WRITE_CMD[38] = len;

  memcpy(&S8K_I2C_WRITE_CMD[6], source, len);

  if (!CardDirect(sizeof(S8K_I2C_WRITE_CMD), S8K_I2C_WRITE_CMD, sizeof(buffer), buffer, status))
    return FALSE;

  return TRUE; 
}